# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: steammessages_clientserver_friends.proto
# plugin: python-betterproto

from dataclasses import dataclass
from typing import List

import betterproto


@dataclass(eq=False, repr=False)
class CMsgClientFriendMsg(betterproto.Message):
    steamid: int = betterproto.fixed64_field(1)
    chat_entry_type: int = betterproto.int32_field(2)
    message: bytes = betterproto.bytes_field(3)
    rtime32_server_timestamp: int = betterproto.fixed32_field(4)
    echo_to_sender: bool = betterproto.bool_field(5)


@dataclass(eq=False, repr=False)
class CMsgClientFriendMsgIncoming(betterproto.Message):
    steamid_from: int = betterproto.fixed64_field(1)
    chat_entry_type: int = betterproto.int32_field(2)
    from_limited_account: bool = betterproto.bool_field(3)
    message: bytes = betterproto.bytes_field(4)
    rtime32_server_timestamp: int = betterproto.fixed32_field(5)


@dataclass(eq=False, repr=False)
class CMsgClientAddFriend(betterproto.Message):
    steamid_to_add: int = betterproto.fixed64_field(1)
    accountname_or_email_to_add: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientAddFriendResponse(betterproto.Message):
    eresult: int = betterproto.int32_field(1)
    steam_id_added: int = betterproto.fixed64_field(2)
    persona_name_added: str = betterproto.string_field(3)


@dataclass(eq=False, repr=False)
class CMsgClientRemoveFriend(betterproto.Message):
    friendid: int = betterproto.fixed64_field(1)


@dataclass(eq=False, repr=False)
class CMsgClientHideFriend(betterproto.Message):
    friendid: int = betterproto.fixed64_field(1)
    hide: bool = betterproto.bool_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientFriendsList(betterproto.Message):
    bincremental: bool = betterproto.bool_field(1)
    friends: List["CMsgClientFriendsListFriend"] = betterproto.message_field(2)
    max_friend_count: int = betterproto.uint32_field(3)
    active_friend_count: int = betterproto.uint32_field(4)
    friends_limit_hit: bool = betterproto.bool_field(5)


@dataclass(eq=False, repr=False)
class CMsgClientFriendsListFriend(betterproto.Message):
    ulfriendid: int = betterproto.fixed64_field(1)
    efriendrelationship: int = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientFriendsGroupsList(betterproto.Message):
    bremoval: bool = betterproto.bool_field(1)
    bincremental: bool = betterproto.bool_field(2)
    friend_groups: List["CMsgClientFriendsGroupsListFriendGroup"] = betterproto.message_field(3)
    memberships: List["CMsgClientFriendsGroupsListFriendGroupsMembership"] = betterproto.message_field(4)


@dataclass(eq=False, repr=False)
class CMsgClientFriendsGroupsListFriendGroup(betterproto.Message):
    n_group_id: int = betterproto.int32_field(1)
    str_group_name: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientFriendsGroupsListFriendGroupsMembership(betterproto.Message):
    ul_steam_id: int = betterproto.fixed64_field(1)
    n_group_id: int = betterproto.int32_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientPlayerNicknameList(betterproto.Message):
    removal: bool = betterproto.bool_field(1)
    incremental: bool = betterproto.bool_field(2)
    nicknames: List["CMsgClientPlayerNicknameListPlayerNickname"] = betterproto.message_field(3)


@dataclass(eq=False, repr=False)
class CMsgClientPlayerNicknameListPlayerNickname(betterproto.Message):
    steamid: int = betterproto.fixed64_field(1)
    nickname: str = betterproto.string_field(3)


@dataclass(eq=False, repr=False)
class CMsgClientSetPlayerNickname(betterproto.Message):
    steamid: int = betterproto.fixed64_field(1)
    nickname: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientSetPlayerNicknameResponse(betterproto.Message):
    eresult: int = betterproto.uint32_field(1)


@dataclass(eq=False, repr=False)
class CMsgClientRequestFriendData(betterproto.Message):
    persona_state_requested: int = betterproto.uint32_field(1)
    friends: List[int] = betterproto.fixed64_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientChangeStatus(betterproto.Message):
    persona_state: int = betterproto.uint32_field(1)
    player_name: str = betterproto.string_field(2)
    is_auto_generated_name: bool = betterproto.bool_field(3)
    high_priority: bool = betterproto.bool_field(4)
    persona_set_by_user: bool = betterproto.bool_field(5)
    persona_state_flags: int = betterproto.uint32_field(6)
    need_persona_response: bool = betterproto.bool_field(7)
    is_client_idle: bool = betterproto.bool_field(8)


@dataclass(eq=False, repr=False)
class CMsgPersonaChangeResponse(betterproto.Message):
    result: int = betterproto.uint32_field(1)
    player_name: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientPersonaState(betterproto.Message):
    status_flags: int = betterproto.uint32_field(1)
    friends: List["CMsgClientPersonaStateFriend"] = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientPersonaStateFriend(betterproto.Message):
    friendid: int = betterproto.fixed64_field(1)
    persona_state: int = betterproto.uint32_field(2)
    game_played_app_id: int = betterproto.uint32_field(3)
    game_server_ip: int = betterproto.uint32_field(4)
    game_server_port: int = betterproto.uint32_field(5)
    persona_state_flags: int = betterproto.uint32_field(6)
    online_session_instances: int = betterproto.uint32_field(7)
    persona_set_by_user: bool = betterproto.bool_field(10)
    player_name: str = betterproto.string_field(15)
    query_port: int = betterproto.uint32_field(20)
    steamid_source: int = betterproto.fixed64_field(25)
    avatar_hash: bytes = betterproto.bytes_field(31)
    last_logoff: int = betterproto.uint32_field(45)
    last_logon: int = betterproto.uint32_field(46)
    last_seen_online: int = betterproto.uint32_field(47)
    clan_rank: int = betterproto.uint32_field(50)
    game_name: str = betterproto.string_field(55)
    gameid: int = betterproto.fixed64_field(56)
    game_data_blob: bytes = betterproto.bytes_field(60)
    clan_data: "CMsgClientPersonaStateFriendClanData" = betterproto.message_field(64)
    clan_tag: str = betterproto.string_field(65)
    rich_presence: List["CMsgClientPersonaStateFriendKv"] = betterproto.message_field(71)
    broadcast_id: int = betterproto.fixed64_field(72)
    game_lobby_id: int = betterproto.fixed64_field(73)
    watching_broadcast_accountid: int = betterproto.uint32_field(74)
    watching_broadcast_appid: int = betterproto.uint32_field(75)
    watching_broadcast_viewers: int = betterproto.uint32_field(76)
    watching_broadcast_title: str = betterproto.string_field(77)


@dataclass(eq=False, repr=False)
class CMsgClientPersonaStateFriendClanData(betterproto.Message):
    ogg_app_id: int = betterproto.uint32_field(1)
    chat_group_id: int = betterproto.uint64_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientPersonaStateFriendKv(betterproto.Message):
    key: str = betterproto.string_field(1)
    value: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientFriendProfileInfo(betterproto.Message):
    steamid_friend: int = betterproto.fixed64_field(1)


@dataclass(eq=False, repr=False)
class CMsgClientFriendProfileInfoResponse(betterproto.Message):
    eresult: int = betterproto.int32_field(1)
    steamid_friend: int = betterproto.fixed64_field(2)
    time_created: int = betterproto.uint32_field(3)
    real_name: str = betterproto.string_field(4)
    city_name: str = betterproto.string_field(5)
    state_name: str = betterproto.string_field(6)
    country_name: str = betterproto.string_field(7)
    headline: str = betterproto.string_field(8)
    summary: str = betterproto.string_field(9)


@dataclass(eq=False, repr=False)
class CMsgClientCreateFriendsGroup(betterproto.Message):
    steamid: int = betterproto.fixed64_field(1)
    groupname: str = betterproto.string_field(2)
    steamid_friends: List[int] = betterproto.fixed64_field(3)


@dataclass(eq=False, repr=False)
class CMsgClientCreateFriendsGroupResponse(betterproto.Message):
    eresult: int = betterproto.uint32_field(1)
    groupid: int = betterproto.int32_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientDeleteFriendsGroup(betterproto.Message):
    steamid: int = betterproto.fixed64_field(1)
    groupid: int = betterproto.int32_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientDeleteFriendsGroupResponse(betterproto.Message):
    eresult: int = betterproto.uint32_field(1)


@dataclass(eq=False, repr=False)
class CMsgClientManageFriendsGroup(betterproto.Message):
    groupid: int = betterproto.int32_field(1)
    groupname: str = betterproto.string_field(2)
    steamid_friends_added: List[int] = betterproto.fixed64_field(3)
    steamid_friends_removed: List[int] = betterproto.fixed64_field(4)


@dataclass(eq=False, repr=False)
class CMsgClientManageFriendsGroupResponse(betterproto.Message):
    eresult: int = betterproto.uint32_field(1)


@dataclass(eq=False, repr=False)
class CMsgClientAddFriendToGroup(betterproto.Message):
    groupid: int = betterproto.int32_field(1)
    steamid: int = betterproto.fixed64_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientAddFriendToGroupResponse(betterproto.Message):
    eresult: int = betterproto.uint32_field(1)


@dataclass(eq=False, repr=False)
class CMsgClientRemoveFriendFromGroup(betterproto.Message):
    groupid: int = betterproto.int32_field(1)
    steamid: int = betterproto.fixed64_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientRemoveFriendFromGroupResponse(betterproto.Message):
    eresult: int = betterproto.uint32_field(1)


@dataclass(eq=False, repr=False)
class CMsgClientGetEmoticonList(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class CMsgClientEmoticonList(betterproto.Message):
    emoticons: List["CMsgClientEmoticonListEmoticon"] = betterproto.message_field(1)
    stickers: List["CMsgClientEmoticonListSticker"] = betterproto.message_field(2)
    effects: List["CMsgClientEmoticonListEffect"] = betterproto.message_field(3)


@dataclass(eq=False, repr=False)
class CMsgClientEmoticonListEmoticon(betterproto.Message):
    name: str = betterproto.string_field(1)
    count: int = betterproto.int32_field(2)
    time_last_used: int = betterproto.uint32_field(3)
    use_count: int = betterproto.uint32_field(4)
    time_received: int = betterproto.uint32_field(5)


@dataclass(eq=False, repr=False)
class CMsgClientEmoticonListSticker(betterproto.Message):
    name: str = betterproto.string_field(1)
    count: int = betterproto.int32_field(2)
    time_received: int = betterproto.uint32_field(3)
    appid: int = betterproto.uint32_field(4)


@dataclass(eq=False, repr=False)
class CMsgClientEmoticonListEffect(betterproto.Message):
    name: str = betterproto.string_field(1)
    count: int = betterproto.int32_field(2)
    time_received: int = betterproto.uint32_field(3)
    infinite_use: bool = betterproto.bool_field(4)
    appid: int = betterproto.uint32_field(5)



# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2015-2020 Rapptz
Copyright (c) 2015 Rossen Georgiev <rossen@rgp.io>
Copyright (c) 2020 James

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

This is a modified version of
https://github.com/Rapptz/discord.py/blob/master/discord/gateway.py
and https://github.com/ValvePython/steam/blob/master/steam/core/cm.py
"""

from __future__ import annotations

import asyncio
import logging
import random
import struct
import sys
import threading
import time
import traceback
from gzip import _GzipReader as GZipReader
from io import BytesIO
from typing import TYPE_CHECKING, Any, Callable, Optional, TypeVar, Union

import aiohttp
import attr
from typing_extensions import Literal

from . import utils
from .enums import EPersonaState, EResult
from .errors import NoCMsFound
from .iterators import AsyncIterator
from .models import Registerable, register
from .protobufs import EMsg, GCMsg, GCMsgProto, Msg, MsgBase, MsgProto
from .protobufs.steammessages_clientserver_2 import CMsgGcClient

if TYPE_CHECKING:
    from .client import Client
    from .enums import EUIMode
    from .game import GameToDict
    from .protobufs.steammessages_base import CMsgMulti
    from .protobufs.steammessages_clientserver_login import CMsgClientLogonResponse
    from .state import ConnectionState, EventParser


__all__ = (
    "ConnectionClosed",
    "CMServerList",
    "SteamWebSocket",
    "return_true",
    "Msgs",
)

log = logging.getLogger(__name__)
Msgs = Union[MsgProto, Msg]
M = TypeVar("M", bound=MsgBase)


def return_true(*_, **__) -> Literal[True]:
    return True


@attr.dataclass(slots=True)
class EventListener:
    emsg: EMsg
    check: Callable[[M], bool]
    future: asyncio.Future[M]


@attr.dataclass(slots=True)
class CMServer:
    url: str
    score: float = 0.0


class ConnectionClosed(Exception):
    def __init__(self, cm: CMServer, cms: CMServerList):
        self.cm = cm
        self.cm_list = cms
        super().__init__(f"Connection to {self.cm.url}, has closed.")


class WebSocketClosure(Exception):
    """An exception to make up for the fact that aiohttp doesn't signal closure."""


class CMServerList(AsyncIterator[CMServer]):
    __slots__ = ("cms", "cell_id", "_state")

    def __init__(self, state: ConnectionState, first_cm_to_try: Optional[CMServer] = None):
        super().__init__(state, None, None, None)
        self.cms: list[CMServer] = []
        self.cell_id = 0
        if first_cm_to_try is not None:
            self.append(first_cm_to_try)

    async def fill(self) -> None:
        if not await self.fetch_servers_from_api():
            raise NoCMsFound("No Community Managers could be found to connect to")

        if not self.cms:
            log.debug("No good servers left. Resetting...")
            self.reset_all()
            return await self.fill()

        random.shuffle(self.cms)
        for cm in await self.ping_cms(self.cms):
            self.append(cm)

    def clear(self) -> None:
        if self.cms:
            log.debug("List cleared.")
        self.cms = []

    async def fetch_servers_from_api(self, cell_id: int = 0) -> bool:
        log.debug("Attempting to fetch servers from the WebAPI")
        self.cell_id = cell_id
        try:
            resp = await self._state.http.get_cm_list(cell_id)
        except Exception as e:
            log.error(f"WebAPI fetch request failed with result: {e!r}")
            return False

        resp = resp["response"]
        if resp["result"] != EResult.OK:
            log.error(
                f"Fetching the CMList failed with: Result: {EResult(resp['result'])!r}. Message: {resp['message']!r}"
            )
            return False

        websockets_list = resp["serverlist_websockets"]
        log.debug(f"Received {len(websockets_list)} servers from WebAPI")

        self.cell_id = cell_id
        self.merge_list(websockets_list)

        return True

    def reset_all(self) -> None:
        log.debug("Marking all CM servers as good.")
        for cm in self.cms:
            cm.score = 0.0

    def merge_list(self, hosts: list[str]) -> None:
        total = len(self.cms)
        urls = [cm.url for cm in self.cms]
        for url in hosts:
            if url not in urls:
                self.cms.append(CMServer(url))
        if len(self.cms) > total:
            log.debug(f"Added {len(self.cms) - total} new CM server addresses.")

    async def ping_cms(self, cms: Optional[list[CMServer]] = None, to_ping: int = 10) -> list[CMServer]:
        cms = self.cms or cms
        best_cms = []
        for cm in cms[:to_ping]:
            # TODO dynamically make sure we get good ones by checking len and stuff
            start = time.perf_counter()
            try:
                resp = await self._state.http._session.get(f"https://{cm.url}/cmping/", timeout=5)
                if resp.status != 200:
                    raise aiohttp.ClientError
                load = resp.headers["X-Steam-CMLoad"]
            except (KeyError, asyncio.TimeoutError, aiohttp.ClientError):
                try:
                    self.cms.remove(cm)
                except ValueError:
                    pass
            else:
                latency = time.perf_counter() - start
                cm.score = (int(load) * 2) + latency
                best_cms.append(cm)
        log.debug("Finished pinging CMs")
        return sorted(best_cms, key=lambda cm: cm.score, reverse=True)


class KeepAliveHandler(threading.Thread):  # ping commands are cool
    def __init__(self, ws: SteamWebSocket, interval: int):
        super().__init__()
        self.ws = ws
        self.interval = interval
        self._main_thread_id = self.ws.thread_id
        self.heartbeat = MsgProto(EMsg.ClientHeartBeat)
        self.heartbeat_timeout = 60
        self.msg = "Keeping websocket alive with heartbeat {}."
        self.block_msg = "Heartbeat blocked for more than {} seconds."
        self.behind_msg = "Can't keep up, websocket is {:.1f} behind."
        self._stop_ev = threading.Event()
        self._last_ack = time.perf_counter()
        self._last_send = time.perf_counter()
        self.latency = float("inf")

    def run(self) -> None:
        while not self._stop_ev.wait(self.interval):
            if self._last_ack + self.heartbeat_timeout < time.perf_counter():
                log.warning(f"Server {self.ws.cm} has stopped responding to the gateway. Closing and restarting.")
                coro = self.ws.handle_close()
                f = asyncio.run_coroutine_threadsafe(coro, loop=self.ws.loop)

                try:
                    f.result()
                except Exception:
                    pass
                finally:
                    return self.stop()

            log.debug(self.msg.format(self.heartbeat))
            coro = self.ws.send_as_proto(self.heartbeat)
            f = asyncio.run_coroutine_threadsafe(coro, loop=self.ws.loop)
            # block until sending is complete
            total = 0
            while True:
                try:
                    f.result(timeout=10)
                except asyncio.TimeoutError:  # alias to concurrent.futures.TimeoutError
                    total += 10
                    try:
                        frame = sys._current_frames()[self._main_thread_id]
                    except KeyError:
                        msg = self.block_msg
                    else:
                        stack = traceback.format_stack(frame)
                        msg = f'{self.block_msg}\nLoop thread traceback (most recent call last):\n{"".join(stack)}'
                    log.warning(msg.format(total))
                except Exception:
                    self.stop()
                else:
                    self.ack()
                    self._last_send = time.perf_counter()
                    break

    def stop(self) -> None:
        self._stop_ev.set()

    def ack(self) -> None:
        self._last_ack = time.perf_counter()
        self.latency = self._last_ack - self._last_send - self.interval
        if self.latency > 10:
            log.warning(self.behind_msg.format(self.latency))


class SteamWebSocket(Registerable):
    parsers: dict[EMsg, EventParser] = {}

    __slots__ = (
        "socket",
        "loop",
        "cm_list",
        "cm",
        "session_id",
        "thread_id",
        "listeners",
        "steam_id",
        "_connection",
        "_dispatch",
        "_current_job_id",
        "_keep_alive",
    )

    def __init__(self, socket: aiohttp.ClientWebSocketResponse, *, loop: asyncio.AbstractEventLoop):
        super().__init__()
        self.socket = socket
        self.loop = loop

        # state stuff
        self._connection: Optional[ConnectionState] = None
        self.cm_list: Optional[CMServerList] = None
        # the keep alive
        self._keep_alive: Optional[KeepAliveHandler] = None
        # an empty dispatcher to prevent crashes
        self._dispatch = lambda *args, **kwargs: None
        self.cm: Optional[CMServer] = None
        self.thread_id = threading.get_ident()

        # ws related stuff
        self.listeners: list[EventListener] = []

        self.session_id = 0
        self.steam_id = 0
        self._current_job_id = 0

    @property
    def latency(self) -> float:
        """:class:`float`: Measures latency between a heartbeat send and the heartbeat interval in seconds."""
        return self._keep_alive.latency

    def wait_for(self, emsg: EMsg, check: Callable[[M], bool] = return_true) -> asyncio.Future[M]:
        future = self.loop.create_future()
        entry = EventListener(emsg=emsg, check=check, future=future)
        self.listeners.append(entry)
        return future

    @classmethod
    async def from_client(
        cls, client: Client, cm: Optional[CMServer] = None, cm_list: Optional[CMServerList] = None
    ) -> SteamWebSocket:
        connection = client._connection
        cm_list = cm_list or CMServerList(connection, cm)
        async for cm in cm_list:
            log.info(f"Attempting to create a websocket connection to: {cm}")
            socket = await client.http.connect_to_cm(cm.url)
            log.debug(f"Connected to {cm}")
            payload = MsgProto(
                EMsg.ClientLogon,
                account_name=client.username,
                web_logon_nonce=client.token,
                client_os_type=4294966596,
                protocol_version=65580,
                chat_mode=2,
                ui_mode=4,
                qos_level=2,
            )
            ws = cls(socket, loop=client.loop)
            # dynamically add attributes needed
            ws._connection = connection
            ws.parsers.update(connection.parsers)
            ws._dispatch = client.dispatch
            ws.steam_id = client.user.id64
            ws.cm = cm
            ws.cm_list = cm_list
            await ws.send_as_proto(payload)  # send the identification message straight away
            ws._dispatch("connect")
            return ws

    async def poll_event(self) -> None:
        try:
            message = await self.socket.receive()
            if message.type is aiohttp.WSMsgType.BINARY and message.data:  # it can sometimes be None/empty
                return await self.receive(message.data)
            if message.type is aiohttp.WSMsgType.ERROR:
                log.debug(f"Received {message}")
                raise message.data
            if message.type in (aiohttp.WSMsgType.CLOSED, aiohttp.WSMsgType.CLOSE, aiohttp.WSMsgType.CLOSING):
                log.debug(f"Received {message}")
                raise WebSocketClosure
            log.debug(f"Dropped unexpected message type: {message}")
        except WebSocketClosure:
            await self.handle_close()

    async def receive(self, message: bytes) -> None:
        self._dispatch("socket_raw_receive", message)

        emsg_value = struct.unpack_from("<I", message)[0]
        emsg = EMsg(utils.clear_proto_bit(emsg_value))

        try:
            msg = MsgProto(emsg, message) if utils.is_proto(emsg_value) else Msg(emsg, message, extended=True)
        except Exception as exc:
            return log.error(f"Failed to deserialize message: {emsg!r}, {message!r}", exc_info=exc)

        if log.isEnabledFor(logging.DEBUG):
            log.debug(f"Socket has received {msg!r} from the websocket.")  # message.__repr__ is an  expensive call

        self._dispatch("socket_receive", msg)

        try:
            event_parser = self.parsers[emsg]
        except KeyError:
            if log.isEnabledFor(logging.DEBUG):
                log.debug(f"Ignoring event {msg!r}")
        else:
            await utils.maybe_coroutine(event_parser, msg)

        # remove the dispatched listener
        removed = []
        for idx, entry in enumerate(self.listeners):
            if entry.emsg != emsg:
                continue

            future = entry.future
            if future.cancelled():
                removed.append(idx)
                continue

            try:
                valid = entry.check(msg)
            except Exception as exc:
                future.set_exception(exc)
                removed.append(idx)
            else:
                if valid:
                    future.set_result(msg)
                    removed.append(idx)

        for idx in reversed(removed):
            del self.listeners[idx]

    async def send(self, data: bytes) -> None:
        self._dispatch("socket_raw_send", data)
        try:
            await self.socket.send_bytes(data=data)
        except ConnectionResetError:
            log.info("Connection closed")
            await self.handle_close()

    async def send_as_proto(self, message: Msgs) -> None:
        message.steam_id = self.steam_id
        message.session_id = self.session_id

        self._dispatch("socket_send", message)
        await self.send(bytes(message))

    async def send_gc_message(self, msg: Union[GCMsgProto, GCMsg]) -> None:  # for ext's to send GC messages
        message: MsgProto[CMsgGcClient] = MsgProto(EMsg.ClientToGC)
        try:
            message.body.appid = message.header.body.routing_appid = self._connection.client.GAME.id
        except AttributeError:
            return utils.warn(f"Attempting to call {self.__class__.__name__}.send_gc_message without a GC Client")
        try:
            if log.isEnabledFor(logging.DEBUG):
                log.debug(f"Sending GC message {msg!r}")
        except Exception:
            log.debug(f"Send GC message {msg.msg}")

        message.body.msgtype = utils.set_proto_bit(msg.msg) if isinstance(msg, GCMsgProto) else msg.msg
        message.body.payload = bytes(msg)
        await self.send_as_proto(message)

    async def close(self, code: int = 1000) -> None:
        message = MsgProto(EMsg.ClientLogOff)
        message.steam_id = self.steam_id
        message.session_id = self.session_id
        await self.socket.close(code=code, message=bytes(message))

    @register(EMsg.ClientLoggedOff)
    async def handle_close(self, _=None) -> None:
        if not self.socket.closed:
            await self.close()
            self.cm_list.queue.pop()  # pop the disconnected cm
        if self._keep_alive is not None:
            self._keep_alive.stop()
            self._keep_alive = None
        log.info(f"Websocket closed, cannot reconnect.")
        raise ConnectionClosed(self.cm, self.cm_list)

    @register(EMsg.ClientLogOnResponse)
    async def handle_logon(self, msg: MsgProto[CMsgClientLogonResponse]) -> None:
        if msg.body.eresult != EResult.OK:
            log.debug(f"Failed to login with result: {EResult(msg.body.eresult)}")
            if msg.body.eresult == EResult.InvalidPassword:
                http = self._connection.http
                await http.logout()
                await http.login(http.username, http.password, shared_secret=http.shared_secret)
            return await self.handle_close()

        self.session_id = msg.session_id
        self.cm_list.cell_id = msg.body.cell_id

        interval = msg.body.out_of_game_heartbeat_seconds
        self._keep_alive = KeepAliveHandler(ws=self, interval=interval)
        self._keep_alive.start()
        log.debug("Heartbeat started.")

        await self.send_um("ChatRoom.GetMyChatRoomGroups#1_Request")
        await self.change_presence(
            games=self._connection._games,
            state=self._connection._state,
            ui_mode=self._connection._ui_mode,
            flags=self._connection._flags,
            force_kick=self._connection._force_kick,
        )
        await self.send_as_proto(MsgProto(EMsg.ClientRequestCommentNotifications))

        log.debug("Logon completed")

    @register(EMsg.Multi)
    async def handle_multi(self, msg: MsgProto[CMsgMulti]) -> None:
        log.debug("Received a multi, unpacking")
        if msg.body.size_unzipped:
            log.debug(f"Decompressing payload ({len(msg.body.message_body)} -> {msg.body.size_unzipped})")
            data = GZipReader(BytesIO(msg.body.message_body)).readall()  # TODO can we be more low level?
            if len(data) != msg.body.size_unzipped:
                return log.info(f"Unzipped size mismatch for multi payload {msg}, discarding")
        else:
            data = msg.body.message_body

        while len(data) > 0:
            size = struct.unpack_from("<I", data)[0]
            await self.receive(data[4 : 4 + size])
            data = data[4 + size :]

    async def send_um(self, name: str, **kwargs: Any) -> int:
        msg = MsgProto(EMsg.ServiceMethodCallFromClient, um_name=name, **kwargs)
        msg.header.body.job_id_source = self._current_job_id = (self._current_job_id + 1) % 10000 or 1
        await self.send_as_proto(msg)
        return msg.header.body.job_id_source

    async def send_um_and_wait(
        self, name: str, check: Optional[Callable[[MsgBase], bool]] = None, timeout: float = 5.0, **kwargs: Any
    ) -> MsgProto:
        job_id = await self.send_um(name, **kwargs)
        check = check or (lambda msg: msg.header.body.job_id_target == job_id)
        return await asyncio.wait_for(self.wait_for(EMsg.ServiceMethodResponse, check=check), timeout=timeout)

    async def change_presence(
        self,
        *,
        games: list[GameToDict],
        state: Optional[EPersonaState],
        flags: int,
        ui_mode: Optional[EUIMode],
        force_kick: bool,
    ) -> None:
        if force_kick:
            kick = MsgProto(EMsg.ClientKickPlayingSession)
            log.debug("Kicking any currently playing sessions")
            await self.send_as_proto(kick)
        if games:
            activity = MsgProto(EMsg.ClientGamesPlayedWithDataBlob, games_played=games)
            log.debug(f"Sending {activity} to change activity")
            await self.send_as_proto(activity)
        if state is not None or flags:
            state = MsgProto(EMsg.ClientChangeStatus, persona_state=state, persona_state_flags=flags)
            log.debug(f"Sending {state} to change state")
            await self.send_as_proto(state)
        if ui_mode is not None:
            ui_mode = MsgProto(EMsg.ClientCurrentUIMode, uimode=ui_mode)
            log.debug(f"Sending {ui_mode} to change UI mode")
            await self.send_as_proto(ui_mode)



# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2020 James

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

import re
import sys
import types
from abc import abstractmethod
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    ForwardRef,
    Generic,
    Iterator,
    NoReturn,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    overload,
)

from typing_extensions import Literal, Protocol, get_args, get_origin, runtime_checkable

from ... import utils
from ...channel import Channel
from ...clan import Clan
from ...errors import HTTPException, InvalidSteamID
from ...game import Game
from ...group import Group
from ...models import FunctionType
from ...user import User
from .errors import BadArgument
from .utils import reload_module_with_TYPE_CHECKING

if TYPE_CHECKING:
    from steam.ext import commands

    from .commands import MC
    from .context import Context

__all__ = (
    "converter_for",
    "Converter",
    "UserConverter",
    "ChannelConverter",
    "ClanConverter",
    "GroupConverter",
    "GameConverter",
    "Default",
    "DefaultAuthor",
    "DefaultChannel",
    "DefaultClan",
    "DefaultGroup",
    "DefaultGame",
    "Greedy",
)

T = TypeVar("T")
Converters = Union[Type["Converter"], "BasicConverter"]
RD = Union[Callable[["MC"], "MC"], "MC"]


class ConverterDict(Dict[type, Tuple[Converters, ...]]):
    def __setitem__(self, key: Any, value: Converters) -> None:
        try:
            old_value = self[key]
        except KeyError:
            super().__setitem__(key, (value,))
        else:
            super().__setitem__(key, old_value + (value,))


class BasicConverter(FunctionType):
    converter_for: T

    def __call__(self, arg: str) -> T:
        ...


CONVERTERS = ConverterDict()


def converter_for(converter_for: T) -> Callable[[MC], MC]:
    """The recommended way to mark a function converter as such.

    Note
    ----
    All of the converters marked with this decorator or derived from :class:`.Converter` can be accessed via
    :attr:`~steam.ext.commands.Bot.converters`.

    Examples
    --------
    .. code-block:: python

        @commands.converter(commands.Command)  # this is the type hint used
        def command_converter(argument: str) -> commands.Command:
            return bot.get_command(argument)

        # then later

        @bot.command
        async def source(ctx, command: commands.Command):  # this then calls command_converter on invocation.
            ...


    Parameters
    ----------
    converter_for: T
        The type annotation the decorated converter should convert for.

    Attributes
    -----------
    converter_for: T
        The class that the converter can be type-hinted to to.
    """

    def decorator(func: MC) -> MC:
        if not isinstance(func, types.FunctionType):
            raise TypeError(f"Excepted a function, received {func.__class__.__name__!r}")
        CONVERTERS[converter_for] = func
        func.converter_for = converter_for
        return func

    return decorator


@runtime_checkable
class Converter(Protocol[T]):
    r"""A custom :class:`typing.Protocol` from which converters can be derived.

    Note
    ----
    All of the converters derived from :class:`Converter` or marked with the :func:`converter_for` decorator can
    be accessed via :attr:`~steam.ext.commands.Bot.converters`.

    Some custom dataclasses from this library can be type-hinted without the need for a custom converter:

        - :class:`~steam.User`.
        - :class:`~steam.abc.Channel`
        - :class:`~steam.Clan`
        - :class:`~steam.Group`
        - :class:`~steam.Game`

    Attributes
    -----------
    converter_for: T
        The class that the converter can be type-hinted to to.

    Examples
    --------

    Builtin: ::

        @bot.command
        async def command(ctx, user: steam.User):
            # this will end up making the user variable a `steam.User` object.

        # invoked as
        # !command 76561198248053954
        # or !command Gobot1234

    A custom converter: ::

        class ImageConverter(commands.Converter[steam.Image]):  # the annotation to typehint to

            async def convert(self, ctx: commands.Context, argument: str) -> steam.Image:
                async with aiohttp.ClientSession() as session:
                    async with session.get(argument) as r:
                        image_bytes = BytesIO(await r.read())
                try:
                    return steam.Image(image_bytes)
                except (TypeError, ValueError):  # failed to convert to an image
                    raise commands.BadArgument("Cannot convert image") from None


        # then later
        @bot.command
        async def set_avatar(ctx: commands.Context, avatar: steam.Image) -> None:
            await bot.user.edit(avatar=avatar)
            await ctx.send("👌")

        # invoked as
        # !set_avatar https://my_image_url.com
    """

    converter_for: T

    @abstractmethod
    async def convert(self, ctx: "commands.Context", argument: str):
        """|coro|
        An abstract method all converters must derive.

        Parameters
        ----------
        ctx: :class:`~steam.ext.commands.Context`
            The context for the invocation.
        argument: :class:`str`
            The argument that is passed from the argument parser.

        Returns
        -------
        T
            The created argument, should be of the same type of :attr:`converter_for`.
        """
        raise NotImplementedError("Derived classes must implement this")

    @classmethod
    @overload
    def register(cls, command: Literal[None] = ...) -> Callable[["MC"], "MC"]:
        ...

    @classmethod
    @overload
    def register(cls, command: MC) -> MC:
        ...

    @classmethod
    def register(cls, command: Optional[RD] = None) -> RD:
        """|maybecallabledeco|
        Register a converter to a specific command.

        Examples
        --------
        .. code-block:: python

            class CustomUserConverter(commands.Converter[steam.User]):
                async def convert(self, ctx: commands.Context, argument: str) -> steam.User:
                    ...

            @bot.command
            @CustomUserConverter.register
            async def is_cool(ctx, user: steam.User):
                ...

        In this example ``is_cool``'s user parameter would be registered to the ``CustomUserConverter`` rather than
        the global :class:`UserConverter`.
        """

        def decorator(command: MC) -> MC:
            is_command = not isinstance(command, types.FunctionType)
            try:
                (command.special_converters if is_command else command.__special_converters__).append(cls)
            except AttributeError:
                if is_command:
                    command.special_converters = [cls]
                else:
                    command.__special_converters__ = [cls]
            return command

        return decorator(command) if command is not None else decorator

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        converter_for = globals().pop("__current_converter", None)
        # the flow for this is __class_getitem__ -> type.__new__ (but that won't do anything) -> __init_subclass__ so
        # this is alright
        if converter_for is None:
            # raise TypeError("Converters should subclass commands.Converter using __class_getitem__")
            utils.warn(
                "Subclassing commands.Converter without arguments is depreciated and is scheduled for removal in V.1"
            )
            CONVERTERS[cls] = cls
        else:
            if isinstance(converter_for, ForwardRef):
                module = sys.modules[cls.__module__]
                reload_module_with_TYPE_CHECKING(module)
                str_value = converter_for.__forward_arg__
                try:
                    converter_for = module.__dict__[str_value]
                except KeyError:
                    raise NameError(f"{str_value!r} was not able to be evaluated to a type") from None
            cls.converter_for = converter_for
            CONVERTERS[converter_for] = cls

    def __class_getitem__(cls, converter_for: ConverterTypes) -> Converter[T]:
        """The entry point for Converters.

        This method is called when :class:`.Converter` is subclassed to handle the argument that was passed as the
        ``converter_for``.
        """
        if isinstance(converter_for, tuple) and len(converter_for) != 1:
            raise TypeError("commands.Converter only accepts one argument")
        annotation = super().__class_getitem__(converter_for)
        globals()["__current_converter"] = get_args(annotation)[0]
        return annotation


class UserConverter(Converter[User]):
    """The converter that is used when the type-hint passed is :class:`~steam.User`.

    Lookup is in the order of:
        - Steam ID
        - Mentions
        - Name
        - URLs
    """

    async def convert(self, ctx: Context, argument: str) -> User:
        try:
            user = await ctx.bot.fetch_user(argument)
        except (InvalidSteamID, HTTPException):
            if argument.startswith("@"):  # probably a mention
                try:
                    account_id = ctx.message.mentions.mention_accountids[0]
                except (IndexError, AttributeError):
                    pass
                else:
                    user = await ctx.bot.fetch_user(account_id)
                    if user is not None and user.id == account_id:
                        ctx.message.mentions.mention_accountids.remove(account_id)
                        return user
            user = utils.find(lambda u: u.name == argument, ctx.bot.users)

            if user is None:
                id64 = await utils.id64_from_url(argument, session=ctx._state.http._session)
                return await self.convert(ctx, id64)
        if user is None:
            raise BadArgument(f'Failed to convert "{argument}" to a Steam user')
        return user


class ChannelConverter(Converter[Channel]):
    """The converter that is used when the type-hint passed is :class:`~steam.Channel`.

    Lookup is in the order of:
        - ID
        - Name
    """

    async def convert(self, ctx: Context, argument: str) -> Channel:
        channel = utils.find(
            lambda c: c.id == int(argument) if argument.isdigit() else lambda c: c.name == argument,
            (ctx.clan or ctx.group).channels,
        )
        if channel is None:
            raise BadArgument(f'Failed to convert "{argument}" to a channel')
        return channel


class ClanConverter(Converter[Clan]):
    """The converter that is used when the type-hint passed is :class:`~steam.Clan`.

    Lookup is in the order of:
        - Steam ID
        - Name
        - URLs
    """

    async def convert(self, ctx: Context, argument: str) -> Clan:
        try:
            clan = await ctx.bot.fetch_clan(argument)
        except (InvalidSteamID, HTTPException):
            clan = utils.find(lambda c: c.name == argument, ctx.bot.clans)
            if clan is None:
                id64 = await utils.id64_from_url(argument, session=ctx._state.http._session)
                return await self.convert(ctx, id64)
        if clan is None:
            raise BadArgument(f'Failed to convert "{argument}" to a Steam clan')
        return clan


class GroupConverter(Converter[Group]):
    """The converter that is used when the type-hint passed is :class:`~steam.Group`.

    Lookup is in the order of:
        - ID
        - Name
    """

    async def convert(self, ctx: Context, argument: str) -> Group:
        try:
            group = ctx.bot.get_group(argument)
        except InvalidSteamID:
            group = utils.find(lambda c: c.name == argument, ctx.bot.groups)
        if group is None:
            raise BadArgument(f'Failed to convert "{argument}" to a Steam group')
        return group


class GameConverter(Converter[Game]):
    """The converter that is used when the type-hint passed is :class:`~steam.Game`.

    If the param is a digit it is assumed that the argument is the :attr:`Game.id` else it is assumed it is the
    :attr:`Game.title`.
    """

    async def convert(self, ctx: Context, argument: str) -> Game:
        return Game(id=int(argument)) if argument.isdigit() else Game(title=argument)


@runtime_checkable
class Default(Protocol):
    """A custom way to specify a default values for commands.

    Examples
    --------
    Builtin: ::

        @bot.command()
        async def info(ctx, user=DefaultAuthor):
            # if no user is passed it will be ctx.author
            ...

    A custom default: ::

        class CurrentCommand(commands.Default):
            async def default(self, ctx: commands.Context) -> commands.Command:
                return ctx.command  # return the current command

        # then later

        @bot.command
        async def source(ctx: commands.Context, command: commands.Command = CurrentCommand):
            # command would now be source
            ...

        # this could also be mixed in with a converter to convert a string to a command.
    """

    @abstractmethod
    async def default(self, ctx: "commands.Context"):
        raise NotImplementedError("Derived classes must to implement this")


class DefaultAuthor(Default):
    """Returns the :attr:`.Context.author`"""

    async def default(self, ctx: Context) -> User:
        return ctx.author


class DefaultChannel(Default):
    """Returns the :attr:`.Context.channel`"""

    async def default(self, ctx: Context) -> Channel:
        return ctx.channel


class DefaultGroup(Default):
    """Returns the :attr:`.Context.group`"""

    async def default(self, ctx: Context) -> Group:
        return ctx.group


class DefaultClan(Default):
    """Returns the :attr:`.Context.clan`"""

    async def default(self, ctx: Context) -> Clan:
        return ctx.clan


class DefaultGame(Default):
    """Returns the :attr:`.Context.author`'s :attr:`~steam.User.game`"""

    async def default(self, ctx: Context) -> Game:
        return ctx.author.game


class Greedy(Generic[T]):
    """
    A custom :class:`typing.Generic` that allows for special greedy command parsing behaviour. It signals to the command
    parser to consume as many arguments as it can until it silently errors reverts the last argument being read and
    then carries on normally.

    Greedy can be mixed in with any normally supported positional or keyword argument type any number of times.

    Example
    -------
    .. code-block:: python

        @bot.command()
        async def test(ctx, numbers: commands.Greedy[int], reason: str):
            await ctx.send(f"numbers: {numbers}, reason: {reason}")

    An invocation of ``"test 1 2 3 4 5 6 hello"`` would pass ``(1, 2, 3, 4, 5, 6)`` to ``numbers`` and ``"hello"`` to
    ``reason``.

    Attributes
    ----------
    converter: T
        The converter the Greedy type holds.

    Note
    ----
    Passing a tuple of length greater than one is shorthand for ``Greedy[Union[converter_tuple]]``.
    """

    converter: T

    def __new__(
        cls, *args: Any, **kwargs: Any
    ) -> NoReturn:  # give a more helpful message than typing._BaseGenericAlias.__call__
        raise TypeError("commands.Greedy cannot be instantiated directly, instead use Greedy[...]")

    def __class_getitem__(cls, converter: GreedyTypes) -> Greedy[T]:
        """The entry point for Greedy types."""
        if isinstance(converter, tuple):
            converter = converter[0] if len(converter) == 1 else Union[converter]
        if not (callable(converter) or isinstance(converter, (Converter, str)) or get_origin(converter) is not None):
            raise TypeError(f"Greedy[...] expects a type or a Converter instance not {converter!r}")

        if converter in INVALID_GREEDY_TYPES:
            raise TypeError(f"Greedy[{converter.__name__}] is invalid")

        origin = get_origin(converter)
        args = get_args(converter)
        if origin is Union:
            for arg in args:
                if arg in INVALID_GREEDY_TYPES:
                    raise TypeError(f"Greedy[{converter!r}] is invalid.")
                if get_origin(arg) is Greedy:  # flatten Greedies similarly to Unions
                    new_args = list(args)
                    idx = new_args.index(arg)
                    new_args.remove(arg)
                    new_args.insert(idx, arg.converter)
                    duplicate_free = []
                    for t in new_args:
                        if t not in duplicate_free:
                            duplicate_free.append(t)
                    return cls.__class_getitem__(tuple(duplicate_free))
                if arg is Greedy:
                    raise TypeError(f"Cannot use un-parametrized Greedy")
        elif origin is Greedy:
            converter = converter.converter

        annotation = super().__class_getitem__(converter)
        annotation.converter = get_args(annotation)[0]
        return annotation

    if TYPE_CHECKING:

        def __iter__(self) -> Iterator[T]:  # make Unions and stuff sort of work
            ...


# fmt: off
ConverterTypes = Union[
    T,
    str,
    Tuple[T, ...],
    Tuple[str, ...],
]
GreedyTypes = Union[
    T,                # a class/type
    str,              # should be a string with a ForwardRef to a class to be evaluated later
    Tuple[T, ...],    # for Greedy[int,] / Greedy[(int,)] to be valid or Greedy[User, int] to be expanded to Union
    Tuple[str, ...],  # same as above two points
    BasicConverter,   # a callable simple converter
    Converter,        # a Converter subclass
]
INVALID_GREEDY_TYPES = (
    str,              # leads to parsing weirdness
    type(None),       # how would this work
)



# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2020 James

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from ...errors import SteamException

if TYPE_CHECKING:
    from inspect import Parameter

    from .commands import Command


__all__ = (
    "CommandError",
    "BadArgument",
    "MissingRequiredArgument",
    "DuplicateKeywordArgument",
    "UnmatchedKeyValuePair",
    "CheckFailure",
    "NotOwner",
    "CommandNotFound",
    "CommandDisabled",
    "CommandOnCooldown",
)


class CommandError(SteamException):
    """Base Exception for errors raised by commands.

    Subclass of :exc:`SteamException`.
    """


class CommandNotFound(CommandError):
    """Exception raised when a command is not found.

    Subclass of :exc:`CommandError`.
    """


class BadArgument(CommandError):
    """Exception raised when a bad argument is passed to a command.

    Subclass of :exc:`CommandError`.
    """


class MissingRequiredArgument(BadArgument):
    """Exception raised when a required argument is not passed to a command.

    Subclass of :exc:`BadArgument`.

    Attributes
    ----------
    param: :class:`inspect.Parameter`
        The argument that is missing.
    """

    def __init__(self, param: Parameter):
        self.param = param
        super().__init__(f"{param.name!r} is a required argument that is missing.")


class DuplicateKeywordArgument(BadArgument):
    """Exception raised when a keyword argument passed would shadow another.

    Subclass of :exc:`BadArgument`.

    Attributes
    ----------
    name: :class:`str`
        The argument that would shadow another.
    """

    def __init__(self, name: str):
        self.name = name
        super().__init__(f"{name!r} shadows another argument.")


class UnmatchedKeyValuePair(BadArgument):
    """Exception raised when an incorrect number of key value pairs are passed to a command to be unpacked.

    Subclass of :exc:`BadArgument`.
    """


class CheckFailure(CommandError):
    """Base Exception raised when a check fails.

    Subclass of :exc:`CommandError`.
    """


class CommandDisabled(CheckFailure):
    """Exception raised when a command is disabled and is attempted to be ran.

    Subclass of :exc:`CheckFailure`.

    Attributes
    -----------
    command: :class:`~steam.ext.commands.Command`
        The command that has been disabled.
    """

    def __init__(self, command: Command):
        self.command = command
        super().__init__(f"{command.name} is currently disabled")


class NotOwner(CheckFailure):
    """Exception raised the user does not own the bot.

    Subclass of :exc:`CheckFailure`.
    """

    def __init__(self):
        super().__init__("You do not own this bot")


class DMChannelOnly(CheckFailure):
    """Exception raised the user does not own the bot.

    Subclass of :exc:`CheckFailure`.
    """

    def __init__(self):
        super().__init__("This command can only be used in DMs")


class CommandOnCooldown(CommandError):
    """Exception raised when a command is still on cooldown.

    Subclass of :exc:`CommandError`.

    Attributes
    ----------
    retry_after: :class:`float`
        The time in seconds at which that the next command can successfully be executed.
    """

    def __init__(self, retry_after: float):
        self.retry_after = retry_after
        super().__init__(f"Command is on cooldown for {retry_after:.2} more seconds")



# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2015 Rossen Georgiev <rossen@rgp.io>
Copyright (c) 2020 James

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

This contains a copy of
https://github.com/ValvePython/steam/blob/master/steam/steamid.py
"""

from __future__ import annotations

import abc
import asyncio
import inspect
import re
from datetime import datetime
from typing import TYPE_CHECKING, Any, Callable, Coroutine, Optional, TypeVar, Union, overload

import attr
from typing_extensions import Final, Protocol, TypeAlias, TypedDict, runtime_checkable

from .badge import FavouriteBadge, UserBadges
from .comment import Comment
from .enums import (
    ECommunityVisibilityState,
    EInstanceFlag,
    EPersonaState,
    EPersonaStateFlag,
    EResult,
    EType,
    ETypeChar,
    EUniverse,
)
from .errors import WSException
from .game import Game
from .iterators import AsyncIterator, CommentsIterator
from .models import Ban, community_route
from .trade import Inventory
from .utils import (
    _INVITE_HEX,
    _INVITE_MAPPING,
    ETypeType,
    EUniverseType,
    InstanceType,
    Intable,
    id64_from_url,
    make_id64,
)

if TYPE_CHECKING:
    from aiohttp import ClientSession

    from .clan import Clan
    from .group import Group
    from .http import StrOrURL
    from .image import Image
    from .protobufs.steammessages_chat import CChatMentions
    from .state import ConnectionState
    from .user import User


__all__ = (
    "SteamID",
    "Message",
    "Channel",
)

C = TypeVar("C", bound="Commentable")


class UserDict(TypedDict):
    steamid: str
    personaname: str
    primaryclanid: str
    profileurl: str
    realname: str
    # enums
    communityvisibilitystate: int
    profilestate: int
    commentpermission: int
    personastate: int
    personastateflags: int
    # avatar
    avatar: str
    avatarmedium: str
    avatarfull: str
    avatarhash: str
    # country stuff
    loccountrycode: str
    locstatecode: int
    loccityid: int
    # game stuff
    gameid: str  # game app id
    gameextrainfo: str  # game name
    # unix timestamps
    timecreated: int
    lastlogoff: int
    # we pass these ourselves
    last_logon: int
    last_seen_online: int


class SteamID(metaclass=abc.ABCMeta):
    """Convert a Steam ID between its various representations.

    Parameters
    ----------
    id: Union[:class:`int`, :class:`str`]
        The ID of the Steam ID, can be an :attr:`id64`, :attr:`id`, :attr:`id2` or an :attr:`id3`.
    type: Union[:class:`.EType`, :class:`int`, :class:`str`]
        The EType for the Steam ID.
    universe: Union[:class:`.EUniverse`, :class:`int`, :class:`str`]
        The EUniverse for the Steam ID.
    instance: :class:`int`
        The instance for the Steam ID.
    """

    __slots__ = ("__BASE",)

    @overload
    def __init__(self):
        ...

    @overload
    def __init__(
        self,
        id: Intable = 0,
        type: Optional[ETypeType] = None,
        universe: Optional[EUniverseType] = None,
        instance: Optional[InstanceType] = None,
    ):
        ...

    def __init__(
        self,
        id: Intable = 0,
        type: Optional[ETypeType] = None,
        universe: Optional[EUniverseType] = None,
        instance: Optional[InstanceType] = None,
    ):
        self.__BASE: Final[int] = make_id64(id, type, universe, instance)

    def __int__(self) -> int:
        return self.__BASE

    def __eq__(self, other: Any) -> bool:
        try:
            return int(self) == int(other)
        except (TypeError, ValueError):
            return NotImplemented

    def __str__(self) -> str:
        return str(self.__BASE)

    def __hash__(self) -> int:
        return hash(self.__BASE)

    def __repr__(self) -> str:
        return f"SteamID(id={self.id}, type={self.type}, universe={self.universe}, instance={self.instance})"

    @property
    def instance(self) -> int:
        """:class:`int`: The instance of the SteamID."""
        return (int(self) >> 32) & 0xfffff

    @property
    def type(self) -> EType:
        """:class:`~steam.EType`: The Steam type of the SteamID."""
        return EType((int(self) >> 52) & 0xf)

    @property
    def universe(self) -> EUniverse:
        """:class:`~steam.EUniverse`: The Steam universe of the SteamID."""
        return EUniverse((int(self) >> 56) & 0xff)

    @property
    def id64(self) -> int:
        """:class:`int`: The SteamID's 64 bit ID."""
        return self.__BASE

    @property
    def id(self) -> int:
        """:class:`int`: The SteamID's 32 bit ID."""
        return int(self) & 0xffffffff

    @property
    def id2(self) -> str:
        """:class:`str`: The SteamID's ID 2.

        e.g ``STEAM_1:0:1234``.
        """
        return f"STEAM_{int(self.universe)}:{self.id % 2}:{self.id >> 1}"

    @property
    def id2_zero(self) -> str:
        """:class:`str`: The SteamID's ID 2 accounted for bugged GoldSrc and Orange Box games.

        Note
        ----
        In these games the accounts :attr:`universe`, ``1`` for :class:`.EType.Public`, should be the ``X`` component of
        ``STEAM_X:0:1234`` however, this was bugged and the value of ``X`` was ``0``.

        e.g ``STEAM_0:0:1234``.
        """
        return self.id2.replace("_1", "_0")

    @property
    def id3(self) -> str:
        """:class:`str`: The SteamID's ID 3.

        e.g ``[U:1:1234]``.
        """
        type_char = ETypeChar(self.type).name
        instance = None

        if self.type in (EType.AnonGameServer, EType.Multiseat):
            instance = self.instance
        elif self.type == EType.Individual:
            if self.instance != 1:
                instance = self.instance
        elif self.type == EType.Chat:
            if self.instance & EInstanceFlag.Clan:
                type_char = "c"
            elif self.instance & EInstanceFlag.Lobby:
                type_char = "L"
            else:
                type_char = "T"

        parts = [type_char, str(self.universe.value), str(self.id)]

        if instance is not None:
            parts.append(str(instance))

        return f"[{':'.join(parts)}]"

    @property
    def invite_code(self) -> Optional[str]:
        """Optional[:class:`str`]: The SteamID's invite code in the s.team invite code format.

        e.g. ``cv-dgb``.
        """
        if self.type == EType.Individual and self.is_valid():
            invite_code = re.sub(f"[{_INVITE_HEX}]", lambda x: _INVITE_MAPPING[x.group()], f"{self.id:x}")
            split_idx = len(invite_code) // 2
            return invite_code if split_idx == 0 else f"{invite_code[:split_idx]}-{invite_code[split_idx:]}"

    @property
    def invite_url(self) -> Optional[str]:
        """Optional[:class:`str`]: The SteamID's full invite code URL.

        e.g ``https://s.team/p/cv-dgb``.
        """
        code = self.invite_code
        return f"https://s.team/p/{code}" if code else None

    @property
    def community_url(self) -> Optional[str]:
        """Optional[:class:`str`]: The SteamID's community url.

        e.g https://steamcommunity.com/profiles/123456789.
        """
        suffix = {
            EType.Individual: "profiles",
            EType.Clan: "gid",
        }
        try:
            return f"https://steamcommunity.com/{suffix[self.type]}/{self.id64}"
        except KeyError:
            return None

    def is_valid(self) -> bool:
        """:class:`bool`: Whether or not the SteamID would be valid."""
        if self.type == EType.Invalid or self.type >= EType.Max:
            return False

        if self.universe == EUniverse.Invalid or self.universe >= EUniverse.Max:
            return False

        if self.type == EType.Individual and (self.id == 0 or self.instance > 4):
            return False

        if self.type == EType.Clan and (self.id == 0 or self.instance != 0):
            return False

        if self.type == EType.GameServer and self.id == 0:
            return False

        if self.type == EType.AnonGameServer and self.id == 0 and self.instance == 0:
            return False

        return True

    @classmethod
    async def from_url(
        cls, url: StrOrURL, session: Optional[ClientSession] = None, timeout: float = 30
    ) -> Optional[SteamID]:
        """|coro|
        A helper function creates a SteamID instance from a Steam community url.

        Parameters
        ----------
        url: Union[:class:`str`, :class:`yarl.URL`]
            The Steam community url to fetch.
        session: Optional[:class:`aiohttp.ClientSession`]
            The session to make the request with. If ``None`` is passed a new one is generated.
        timeout: :class:`float`
            How long to wait for a response before returning ``None``.

        Returns
        -------
        Optional[:class:`SteamID`]
            :class:`SteamID` instance or ``None``.
        """
        id64 = await id64_from_url(url, session, timeout)
        return cls(id64) if id64 else None


class Commentable(SteamID):
    """A mixin that implements commenting functionality"""

    __slots__ = (
        "comment_path",
        "_state",
        "__weakref__",
    )

    comment_path: Final[str]  # noqa  # type: ignore

    def __init_subclass__(cls, comment_path: str = "Profile") -> None:
        cls.comment_path: Final[str] = comment_path  # noqa  # type: ignore

    def copy(self: C) -> C:
        cls = self.__class__
        commentable = cls.__new__(cls)
        for name, attr in inspect.getmembers(self):
            try:
                setattr(commentable, name, attr)
            except (AttributeError, TypeError):
                pass
        return commentable

    __copy__ = copy

    async def comment(self, content: str) -> Comment:
        """|coro|
        Post a comment to a profile.

        Parameters
        -----------
        content: :class:`str`
            The message to add to the profile.

        Returns
        -------
        :class:`~steam.Comment`
            The created comment.
        """
        resp = await self._state.http.post_comment(self.id64, self.comment_path, content)
        id = int(re.findall(r'id="comment_(\d+)"', resp["comments_html"])[0])
        timestamp = datetime.utcfromtimestamp(resp["timelastpost"])
        comment = Comment(
            state=self._state, id=id, owner=self, timestamp=timestamp, content=content, author=self._state.client.user
        )
        self._state.dispatch("comment", comment)
        return comment

    def comments(
        self, limit: Optional[int] = None, before: Optional[datetime] = None, after: Optional[datetime] = None
    ) -> CommentsIterator:
        """An :class:`~steam.iterators.AsyncIterator` for accessing a profile's :class:`~steam.Comment` objects.

        Examples
        ---------

        Usage: ::

            async for comment in commentable.comments(limit=10):
                print('Author:', comment.author, 'Said:', comment.content)

        Flattening into a list: ::

            comments = await commentable.comments(limit=50).flatten()
            # comments is now a list of Comment

        All parameters are optional.

        Parameters
        ----------
        limit: Optional[:class:`int`]
            The maximum number of comments to search through.
            Default is ``None`` which will fetch the clan's entire comments section.
        before: Optional[:class:`datetime.datetime`]
            A time to search for comments before.
        after: Optional[:class:`datetime.datetime`]
            A time to search for comments after.

        Yields
        ---------
        :class:`~steam.Comment`
            The comment with the comment information parsed.
        """
        return CommentsIterator(state=self._state, owner=self, limit=limit, before=before, after=after)


class BaseUser(Commentable):
    """An ABC that details the common operations on a Steam user.
    The following classes implement this ABC:

        - :class:`~steam.User`
        - :class:`~steam.ClientUser`

    .. container:: operations

        .. describe:: x == y

            Checks if two users are equal.

        .. describe:: str(x)

            Returns the user's name.

    Attributes
    ----------
    name: :class:`str`
        The user's username.
    state: :class:`~steam.EPersonaState`
        The current persona state of the account (e.g. LookingToTrade).
    game: Optional[:class:`~steam.Game`]
        The Game instance attached to the user. Is ``None`` if the user isn't in a game or one that is recognised by the
        api.
    primary_clan: Optional[:class:`SteamID`]
        The primary clan the User displays on their profile.
    avatar_url: :class:`str`
        The avatar url of the user. Uses the large (184x184 px) image url.
    real_name: Optional[:class:`str`]
        The user's real name defined by them. Could be ``None``.
    created_at: Optional[:class:`datetime.datetime`]
        The time at which the user's account was created. Could be ``None``.
    last_logoff: Optional[:class:`datetime.datetime`]
        The last time the user logged into steam. Could be None (e.g. if they are currently online).
    country: Optional[:class:`str`]
        The country code of the account. Could be ``None``.
    flags: list[:class:`~steam.EPersonaStateFlag`]
        The persona state flags of the account.
    """

    __slots__ = (
        "name",
        "game",
        "state",
        "flags",
        "country",
        "primary_clan",
        "trade_url",
        "real_name",
        "avatar_url",
        "last_seen_online",
        "created_at",
        "last_logoff",
        "last_logon",
        "privacy_state",
        "favourite_badge",
        "_is_commentable",
        "_setup_profile",
        "_level",
    )

    def __init__(self, state: ConnectionState, data: UserDict):
        super().__init__(data["steamid"])
        self._state = state
        self.name: Optional[str] = None
        self.real_name: Optional[str] = None
        self.avatar_url: Optional[str] = None
        self.primary_clan: Optional[SteamID] = None
        self.country: Optional[str] = None
        self.created_at: Optional[datetime] = None
        self.last_logoff: Optional[datetime] = None
        self.last_logon: Optional[datetime] = None
        self.last_seen_online: Optional[datetime] = None
        self.game: Optional[Game] = None
        self.state: Optional[EPersonaState] = None
        self.flags: list[EPersonaStateFlag] = []
        self.privacy_state: Optional[ECommunityVisibilityState] = None
        self._update(data)

    def _update(self, data: UserDict) -> None:
        self.name = data["personaname"]
        self.real_name = data.get("realname") or self.real_name
        self.avatar_url = data.get("avatarfull") or self.avatar_url
        self.trade_url = community_route(f"tradeoffer/new/?partner={self.id}")

        self.primary_clan = SteamID(data["primaryclanid"]) if "primaryclanid" in data else self.primary_clan
        self.country = data.get("loccountrycode") or self.country
        self.created_at = datetime.utcfromtimestamp(data["timecreated"]) if "timecreated" in data else self.created_at
        self.last_logoff = datetime.utcfromtimestamp(data["lastlogoff"]) if "lastlogoff" in data else self.last_logoff
        self.last_logon = datetime.utcfromtimestamp(data["last_logon"]) if "last_logon" in data else self.last_logon
        self.last_seen_online = (
            datetime.utcfromtimestamp(data["last_seen_online"]) if "last_seen_online" in data else self.last_seen_online
        )
        self.game = Game(title=data.get("gameextrainfo"), id=data["gameid"]) if "gameid" in data else self.game
        self.state = EPersonaState(data.get("personastate", 0)) or self.state
        self.flags = EPersonaStateFlag.components(data.get("personastateflags", 0)) or self.flags
        self.privacy_state = ECommunityVisibilityState(data.get("communityvisibilitystate", 0))
        self._is_commentable = bool(data.get("commentpermission"))
        self._setup_profile = bool(data.get("profilestate"))

    def __repr__(self) -> str:
        attrs = ("name", "state", "id", "type", "universe", "instance")
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        return f"<{self.__class__.__name__} {' '.join(resolved)}>"

    def __str__(self) -> str:
        return self.name

    @property
    def mention(self) -> str:
        """:class:`str`: The string used to mention the user."""
        return f"[mention={self.id}]@{self.name}[/mention]"

    async def inventory(self, game: Game) -> Inventory:
        """|coro|
        Fetch an :class:`User`'s :class:`~steam.Inventory` for trading.

        Parameters
        -----------
        game: :class:`~steam.Game`
            The game to fetch the inventory for.

        Raises
        ------
        :exc:`~steam.Forbidden`
            The user's inventory is private.

        Returns
        -------
        :class:`Inventory`
            The user's inventory.
        """
        resp = await self._state.http.get_user_inventory(self.id64, game.id, game.context_id)
        return Inventory(state=self._state, data=resp, owner=self)

    async def friends(self) -> list[User]:
        """|coro|
        Fetch the list of :class:`~steam.User`'s friends from the API.

        Returns
        -------
        list[:class:`~steam.User`]
            The list of user's friends from the API.
        """
        friends = await self._state.http.get_friends(self.id64)
        return [self._state._store_user(friend) for friend in friends]

    async def games(self) -> list[Game]:
        """|coro|
        Fetches the :class:`~steam.Game` objects the :class:`User` owns from the API.

        Returns
        -------
        list[:class:`~steam.Game`]
            The list of game objects from the API.
        """
        data = await self._state.http.get_user_games(self.id64)
        games = data["response"].get("games", [])
        return [Game._from_api(game) for game in games]

    async def clans(self) -> list[Clan]:
        """|coro|
        Fetches a list of the :class:`User`'s :class:`~steam.Clan`
        objects the :class:`User` is in from the API.

        Returns
        -------
        list[:class:`~steam.Clan`]
            The user's clans.
        """
        clans = []

        async def getter(gid: int) -> None:
            try:
                clan = await self._state.client.fetch_clan(gid)
            except WSException as exc:
                if exc.code == EResult.RateLimitExceeded:
                    await asyncio.sleep(20)
                    await getter(gid)
            else:
                clans.append(clan)

        resp = await self._state.http.get_user_clans(self.id64)
        for clan in resp["response"]["groups"]:
            await getter(int(clan["gid"]))
        return clans

    async def bans(self) -> Ban:
        """|coro|
        Fetches the :class:`User`'s :class:`~steam.Ban` objects.

        Returns
        -------
        :class:`~steam.Ban`
            The user's bans.
        """
        resp = await self._state.http.get_user_bans(self.id64)
        resp = resp["players"][0]
        resp["EconomyBan"] = resp["EconomyBan"] != "none"
        return Ban(data=resp)

    async def badges(self) -> UserBadges:
        """|coro|
        Fetches the :class:`User`'s :class:`~steam.UserBadges` objects.

        Returns
        -------
        :class:`~steam.UserBadges`
            The user's badges.
        """
        resp = await self._state.http.get_user_badges(self.id64)
        return UserBadges(data=resp["response"])

    async def level(self) -> int:
        """|coro|
        Fetches the :class:`User`'s level.

        Returns
        -------
        :class:`int`
            The user's level.
        """
        if self._state.http.api_key is not None:
            resp = await self._state.http.get_user_level(self.id64)
            return resp["response"]["player_level"]
        return self._level

    def is_commentable(self) -> bool:
        """:class:`bool`: Specifies if the user's account is able to be commented on."""
        return self._is_commentable

    def is_private(self) -> bool:
        """:class:`bool`: Specifies if the user has a private profile."""
        return self.privacy_state == ECommunityVisibilityState.Private

    def has_setup_profile(self) -> bool:
        """:class:`bool`: Specifies if the user has a setup their profile."""
        return self._setup_profile

    async def is_banned(self) -> bool:
        """|coro|
        Specifies if the user is banned from any part of Steam.

        This is equivalent to: ::

            bans = await user.bans()
            bans.is_banned()

        Returns
        -------
        :class:`bool`
            Whether or not the user is banned.
        """
        bans = await self.bans()
        return bans.is_banned()

    @classmethod
    def _patch_without_api(cls):
        import functools

        def __init__(self, state: ConnectionState, data: dict) -> None:
            super().__init__(data["steamid"])
            self._state = state
            self.name = data["persona_name"]
            self.avatar_url = data.get("avatar_url") or self.avatar_url

            self.real_name = NotImplemented
            self.trade_url = NotImplemented
            self.primary_clan = NotImplemented
            self.country = NotImplemented
            self.created_at = NotImplemented
            self.last_logoff = NotImplemented
            self.last_logon = NotImplemented
            self.last_seen_online = NotImplemented
            self.game = NotImplemented
            self.state = NotImplemented
            self.flags = NotImplemented
            self.privacy_state = NotImplemented
            self._is_commentable = NotImplemented
            self._setup_profile = NotImplemented

            try:
                favourite_badge = data["favorite_badge"]
            except KeyError:
                self.favourite_badge = None
            else:
                icon_url = favourite_badge.pop("icon")
                self.favourite_badge = FavouriteBadge(**favourite_badge, icon_url=icon_url)

            self._level = data["level"]

        def __repr__(self) -> str:
            attrs = ("name", "id", "type", "universe", "instance")
            resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
            return f"<{self.__class__.__name__} {' '.join(resolved)}>"

        setattr(cls, "__init__", __init__)
        setattr(cls, "__repr__", __repr__)

        def not_implemented(function: str):
            @functools.wraps(getattr(cls, function))
            async def wrapped(*_, **__) -> None:
                raise NotImplementedError(
                    f"Accounts without an API key cannot use User.{function}, this is a Steam limitation not a library "
                    f"limitation, sorry."
                )

            setattr(cls, function, wrapped)

        not_implemented("friends")
        not_implemented("badges")
        not_implemented("is_commentable")
        not_implemented("has_setup_profile")
        not_implemented("is_private")
        not_implemented("is_banned")


_EndPointReturnType: TypeAlias = "tuple[Union[tuple[int, ...], int], Callable[..., Coroutine[None, None, None]]]"


@runtime_checkable
class Messageable(Protocol):
    """An ABC that details the common operations on a Steam message.
    The following classes implement this ABC:

        - :class:`~steam.User`
        - :class:`~steam.ClanChannel`
        - :class:`~steam.GroupChannel`
        - :class:`~steam.DMChannel`
    """

    __slots__ = ()

    @abc.abstractmethod
    def _get_message_endpoint(self) -> _EndPointReturnType:
        raise NotImplementedError

    @abc.abstractmethod
    def _get_image_endpoint(self) -> _EndPointReturnType:
        raise NotImplementedError

    async def send(self, content: Optional[str] = None, image: Optional[Image] = None) -> None:
        """|coro|
        Send a message to a certain destination.

        Parameters
        ----------
        content: Optional[:class:`str`]
            The content of the message to send.
        image: Optional[:class:`.Image`]
            The image to send to the user.

        Raises
        ------
        :exc:`~steam.HTTPException`
            Sending the message failed.
        :exc:`~steam.Forbidden`
            You do not have permission to send the message.
        """
        if content is not None:
            destination, message_func = self._get_message_endpoint()
            await message_func(destination, str(content))
        if image is not None:
            destination, image_func = self._get_image_endpoint()
            await image_func(destination, image)


@attr.dataclass(slots=True)
class Channel(Messageable):
    _state: ConnectionState
    clan: Optional[Clan] = None
    group: Optional[Group] = None

    @abc.abstractmethod
    def history(
        self,
        limit: Optional[int] = 100,
        before: Optional[datetime] = None,
        after: Optional[datetime] = None,
    ) -> AsyncIterator["Message"]:
        """An :class:`~steam.iterators.AsyncIterator` for accessing a :class:`steam.Channel`'s :class:`steam.Message`
        objects.

        Examples
        --------

        Usage: ::

            async for message in channel.history(limit=10):
                print('Author:', message.author, 'Said:', message.content)

        Flattening into a list: ::

            messages = await channel.history(limit=50).flatten()
            # messages is now a list of Message

        All parameters are optional.

        Parameters
        ----------
        limit: Optional[:class:`int`]
            The maximum number of trades to search through.
            Setting this to ``None`` will fetch all of the channel's messages, but this will be a very slow operation.
        before: Optional[:class:`datetime.datetime`]
            A time to search for messages before.
        after: Optional[:class:`datetime.datetime`]
            A time to search for messages after.

        Yields
        ------
        :class:`~steam.Message`
        """
        raise NotImplementedError


def _clean_up_content(content: str) -> str:  # steam does weird stuff with content
    return content.replace(r"\[", "[").replace("\\\\", "\\")


class Message:
    """Represents a message from a :class:`~steam.User`. This is a base class from which all messages inherit.

    The following classes implement this:

        - :class:`~steam.UserMessage`
        - :class:`~steam.GroupMessage`
        - :class:`~steam.ClanMessage`

    Attributes
    ----------
    channel: :class:`Channel`
        The channel the message was sent in.
    content: :class:`str`
        The message's content.

        Note
        ----
        This is **not** what you will see in the steam client see :attr:`clean_content` for that.
    clean_content: :class:`str`
        The message's clean content without bbcode.
    author: :class:`~steam.User`
        The message's author.
    created_at: :class:`datetime.datetime`
        The time the message was sent at.
    group: Optional[:class:`~steam.Group`]
        The group the message was sent in. Will be ``None`` if the message wasn't sent in a :class:`~steam.Group`.
    clan: Optional[:class:`~steam.Clan`]
        The clan the message was sent in. Will be ``None`` if the message wasn't sent in a :class:`~steam.Clan`.
    """

    __slots__ = (
        "author",
        "content",
        "channel",
        "clean_content",
        "created_at",
        "group",
        "clan",
        "mentions",
        "_state",
    )
    author: User
    created_at: datetime

    def __init__(self, channel: Channel, proto: Any):
        self._state: ConnectionState = channel._state
        self.channel = channel
        self.group: Optional[Group] = channel.group
        self.clan: Optional[Clan] = channel.clan
        self.content: str = _clean_up_content(proto.message)
        self.clean_content: str = getattr(proto, "message_no_bbcode", None) or self.content
        self.mentions: Optional[CChatMentions] = getattr(proto, "mentions", None)

    def __repr__(self) -> str:
        attrs = ("author", "channel")
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        return f"<{self.__class__.__name__} {' '.join(resolved)}>"



# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2016 Michał Bukowski gigibukson@gmail.com
Copyright (c) 2017 Nate the great
Copyright (c) 2020 James

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

This contains a copy of
https://github.com/bukson/steampy/blob/master/steampy/guard.py
and https://github.com/Zwork101/steam-trade/blob/master/pytrade/confirmations.py
with extra doc-strings and performance improvements.
"""

from __future__ import annotations

import base64
import hmac
import struct
from hashlib import sha1
from time import time
from typing import TYPE_CHECKING, Any, Optional

from .errors import ConfirmationError
from .models import community_route

if TYPE_CHECKING:
    from .state import ConnectionState


__all__ = (
    "generate_one_time_code",
    "generate_confirmation_code",
    "generate_device_id",
    "Confirmation",
)


def generate_one_time_code(shared_secret: str, timestamp: Optional[int] = None) -> str:
    """Generate a Steam Guard code for signing in.

    Parameters
    -----------
    shared_secret: :class:`str`
        Identity secret from steam guard.
    timestamp: Optional[:class:`int`]
        The unix timestamp to generate the key for.

    Returns
    --------
    :class:`str`
        The desired 2FA code for the timestamp.
    """
    timestamp = timestamp or int(time())
    time_buffer = struct.pack(">Q", timestamp // 30)  # pack as Big endian, uint64
    time_hmac = hmac.new(base64.b64decode(shared_secret), time_buffer, digestmod=sha1).digest()
    begin = ord(time_hmac[19:20]) & 0xf
    full_code = struct.unpack(">I", time_hmac[begin : begin + 4])[0] & 0x7fffffff  # unpack as Big endian uint32
    chars = "23456789BCDFGHJKMNPQRTVWXY"
    code = []
    for _ in range(5):
        full_code, i = divmod(full_code, len(chars))
        code.append(chars[i])
    return "".join(code)


def generate_confirmation_code(identity_secret: str, tag: str, timestamp: Optional[int] = None) -> str:
    """Generate a trade confirmation code.

    Parameters
    -----------
    identity_secret: :class:`str`
        Identity secret from steam guard.
    tag: :class:`str`
        Tag to encode to.
    timestamp: Optional[:class:`int`]
        The time to generate the key for.

    Returns
    --------
    :class:`str`
        Confirmation code for the set timestamp.
    """
    timestamp = timestamp or int(time())
    buffer = struct.pack(">Q", timestamp) + tag.encode("ascii")
    return base64.b64encode(hmac.new(base64.b64decode(identity_secret), buffer, digestmod=sha1).digest()).decode()


def generate_device_id(user_id64: str) -> str:
    """
    Parameters
    -----------
    user_id64: :class:`str`
        The 64 bit steam id to generate the device id for.

    Returns
    --------
    :class:`str`
        The device id.
    """
    # it works, however it's different that one generated from mobile app

    hexed_steam_id = sha1(user_id64.encode("ascii")).hexdigest()
    partial_id = [
        hexed_steam_id[:8],
        hexed_steam_id[8:12],
        hexed_steam_id[12:16],
        hexed_steam_id[16:20],
        hexed_steam_id[20:32],
    ]
    return f'android:{"-".join(partial_id)}'


class Confirmation:
    __slots__ = (
        "_state",
        "id",
        "data_conf_id",
        "data_key",
        "tag",
        "trade_id",
    )

    def __init__(
        self,
        state: ConnectionState,
        confirmation_id: int,
        data_conf_id: int,
        data_key: str,
        trade_id: int,
        tag: str,
    ):
        self._state = state
        self.id = confirmation_id
        self.data_conf_id = data_conf_id
        self.data_key = data_key
        self.tag = tag
        self.trade_id = trade_id  # this isn't really always the trade ID, but for our purposes this is fine

    def __repr__(self) -> str:
        return f"<Confirmation id={self.id!r} trade_id={self.trade_id}>"

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, Confirmation) and self.trade_id == other.trade_id and self.id == other.id

    def _confirm_params(self, tag: str) -> dict:
        timestamp = int(time())
        return {
            "p": self._state._device_id,
            "a": self._state._id64,
            "k": self._state._generate_confirmation(tag, timestamp),
            "t": timestamp,
            "m": "android",
            "tag": tag,
        }

    def _assert_valid(self, resp: dict) -> None:
        if not resp.get("success", False):
            self._state._confirmations_to_ignore.append(self.trade_id)
            raise ConfirmationError

    async def confirm(self) -> None:
        params = self._confirm_params("allow")
        params["op"] = "allow"
        params["cid"] = self.data_conf_id
        params["ck"] = self.data_key
        resp = await self._state.request("GET", community_route("mobileconf/ajaxop"), params=params)
        self._assert_valid(resp)

    async def cancel(self) -> None:
        params = self._confirm_params("cancel")
        params["op"] = "cancel"
        params["cid"] = self.data_conf_id
        params["ck"] = self.data_key
        resp = await self._state.request("GET", community_route("mobileconf/ajaxop"), params=params)
        self._assert_valid(resp)

    async def details(self) -> dict:
        params = self._confirm_params(self.tag)
        resp = await self._state.request("GET", community_route(f"mobileconf/details/{self.id}"), params=params)
        self._assert_valid(resp)
        return resp["html"]



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: steammessages_clientserver_login.proto
# plugin: python-betterproto

from dataclasses import dataclass

import betterproto

from .steammessages_base import CMsgIpAddress


@dataclass(eq=False, repr=False)
class CMsgClientHeartBeat(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class CMsgClientServerTimestampRequest(betterproto.Message):
    client_request_timestamp: int = betterproto.uint64_field(1)


@dataclass(eq=False, repr=False)
class CMsgClientServerTimestampResponse(betterproto.Message):
    client_request_timestamp: int = betterproto.uint64_field(1)
    server_timestamp_ms: int = betterproto.uint64_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientSecret(betterproto.Message):
    version: int = betterproto.uint32_field(1)
    appid: int = betterproto.uint32_field(2)
    deviceid: int = betterproto.uint32_field(3)
    nonce: int = betterproto.fixed64_field(4)
    hmac: bytes = betterproto.bytes_field(5)


@dataclass(eq=False, repr=False)
class CMsgClientLogon(betterproto.Message):
    protocol_version: int = betterproto.uint32_field(1)
    deprecated_obfustucated_private_ip: int = betterproto.uint32_field(2)
    cell_id: int = betterproto.uint32_field(3)
    last_session_id: int = betterproto.uint32_field(4)
    client_package_version: int = betterproto.uint32_field(5)
    client_language: str = betterproto.string_field(6)
    client_os_type: int = betterproto.uint32_field(7)
    should_remember_password: bool = betterproto.bool_field(8)
    wine_version: str = betterproto.string_field(9)
    deprecated_10: int = betterproto.uint32_field(10)
    obfuscated_private_ip: "CMsgIpAddress" = betterproto.message_field(11)
    deprecated_public_ip: int = betterproto.uint32_field(20)
    qos_level: int = betterproto.uint32_field(21)
    client_supplied_steam_id: int = betterproto.fixed64_field(22)
    public_ip: "CMsgIpAddress" = betterproto.message_field(23)
    machine_id: bytes = betterproto.bytes_field(30)
    launcher_type: int = betterproto.uint32_field(31)
    ui_mode: int = betterproto.uint32_field(32)
    chat_mode: int = betterproto.uint32_field(33)
    steam2_auth_ticket: bytes = betterproto.bytes_field(41)
    email_address: str = betterproto.string_field(42)
    rtime32_account_creation: int = betterproto.fixed32_field(43)
    account_name: str = betterproto.string_field(50)
    password: str = betterproto.string_field(51)
    game_server_token: str = betterproto.string_field(52)
    login_key: str = betterproto.string_field(60)
    was_converted_deprecated_msg: bool = betterproto.bool_field(70)
    anon_user_target_account_name: str = betterproto.string_field(80)
    resolved_user_steam_id: int = betterproto.fixed64_field(81)
    eresult_sentryfile: int = betterproto.int32_field(82)
    sha_sentryfile: bytes = betterproto.bytes_field(83)
    auth_code: str = betterproto.string_field(84)
    otp_type: int = betterproto.int32_field(85)
    otp_value: int = betterproto.uint32_field(86)
    otp_identifier: str = betterproto.string_field(87)
    steam2_ticket_request: bool = betterproto.bool_field(88)
    sony_psn_ticket: bytes = betterproto.bytes_field(90)
    sony_psn_service_id: str = betterproto.string_field(91)
    create_new_psn_linked_account_if_needed: bool = betterproto.bool_field(92)
    sony_psn_name: str = betterproto.string_field(93)
    game_server_app_id: int = betterproto.int32_field(94)
    steamguard_dont_remember_computer: bool = betterproto.bool_field(95)
    machine_name: str = betterproto.string_field(96)
    machine_name_userchosen: str = betterproto.string_field(97)
    country_override: str = betterproto.string_field(98)
    is_steam_box: bool = betterproto.bool_field(99)
    client_instance_id: int = betterproto.uint64_field(100)
    two_factor_code: str = betterproto.string_field(101)
    supports_rate_limit_response: bool = betterproto.bool_field(102)
    web_logon_nonce: str = betterproto.string_field(103)
    priority_reason: int = betterproto.int32_field(104)
    embedded_client_secret: "CMsgClientSecret" = betterproto.message_field(105)
    disable_partner_autogrants: bool = betterproto.bool_field(106)


@dataclass(eq=False, repr=False)
class CMsgClientLogonResponse(betterproto.Message):
    eresult: int = betterproto.int32_field(1)
    out_of_game_heartbeat_seconds: int = betterproto.int32_field(2)
    in_game_heartbeat_seconds: int = betterproto.int32_field(3)
    deprecated_public_ip: int = betterproto.uint32_field(4)
    rtime32_server_time: int = betterproto.fixed32_field(5)
    account_flags: int = betterproto.uint32_field(6)
    cell_id: int = betterproto.uint32_field(7)
    email_domain: str = betterproto.string_field(8)
    steam2_ticket: bytes = betterproto.bytes_field(9)
    eresult_extended: int = betterproto.int32_field(10)
    webapi_authenticate_user_nonce: str = betterproto.string_field(11)
    cell_id_ping_threshold: int = betterproto.uint32_field(12)
    deprecated_use_pics: bool = betterproto.bool_field(13)
    vanity_url: str = betterproto.string_field(14)
    public_ip: "CMsgIpAddress" = betterproto.message_field(15)
    client_supplied_steamid: int = betterproto.fixed64_field(20)
    ip_country_code: str = betterproto.string_field(21)
    parental_settings: bytes = betterproto.bytes_field(22)
    parental_setting_signature: bytes = betterproto.bytes_field(23)
    count_loginfailures_to_migrate: int = betterproto.int32_field(24)
    count_disconnects_to_migrate: int = betterproto.int32_field(25)
    ogs_data_report_time_window: int = betterproto.int32_field(26)
    client_instance_id: int = betterproto.uint64_field(27)
    force_client_update_check: bool = betterproto.bool_field(28)


@dataclass(eq=False, repr=False)
class CMsgClientRequestWebApiAuthenticateUserNonce(betterproto.Message):
    token_type: int = betterproto.int32_field(1)


@dataclass(eq=False, repr=False)
class CMsgClientRequestWebApiAuthenticateUserNonceResponse(betterproto.Message):
    eresult: int = betterproto.int32_field(1)
    webapi_authenticate_user_nonce: str = betterproto.string_field(11)
    token_type: int = betterproto.int32_field(3)


@dataclass(eq=False, repr=False)
class CMsgClientLogOff(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class CMsgClientLoggedOff(betterproto.Message):
    eresult: int = betterproto.int32_field(1)


@dataclass(eq=False, repr=False)
class CMsgClientNewLoginKey(betterproto.Message):
    unique_id: int = betterproto.uint32_field(1)
    login_key: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class CMsgClientNewLoginKeyAccepted(betterproto.Message):
    unique_id: int = betterproto.uint32_field(1)


@dataclass(eq=False, repr=False)
class CMsgClientAccountInfo(betterproto.Message):
    persona_name: str = betterproto.string_field(1)
    ip_country: str = betterproto.string_field(2)
    count_authed_computers: int = betterproto.int32_field(5)
    account_flags: int = betterproto.uint32_field(7)
    facebook_id: int = betterproto.uint64_field(8)
    facebook_name: str = betterproto.string_field(9)
    steamguard_notify_newmachines: bool = betterproto.bool_field(14)
    steamguard_machine_name_user_chosen: str = betterproto.string_field(15)
    is_phone_verified: bool = betterproto.bool_field(16)
    two_factor_state: int = betterproto.uint32_field(17)
    is_phone_identifying: bool = betterproto.bool_field(18)
    is_phone_needing_reverify: bool = betterproto.bool_field(19)


@dataclass(eq=False, repr=False)
class CMsgClientChallengeRequest(betterproto.Message):
    steamid: int = betterproto.fixed64_field(1)


@dataclass(eq=False, repr=False)
class CMsgClientChallengeResponse(betterproto.Message):
    challenge: int = betterproto.fixed64_field(1)



# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2020 James

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from .abc import SteamID
from .channel import GroupChannel
from .role import Role

if TYPE_CHECKING:
    from .protobufs.steammessages_chat import CChatRoomGetChatRoomGroupSummaryResponse as GroupProto
    from .state import ConnectionState
    from .user import User


__all__ = ("Group",)


class Group(SteamID):
    """Represents a Steam group.

    Attributes
    ----------
    name: Optional[:class:`str`]
        The name of the group, could be ``None``.
    owner: :class:`~steam.abc.BaseUser`
        The owner of the group.
    top_members: list[:class:`~steam.abc.BaseUser`]
        A list of the group's top members.
    active_member_count: :class:`int`
        The group's active member count.
    roles: list[:class:`~steam.Role`]
        A list of the group's roles.
    default_role: :class:`~steam.Role`
        The group's default role.
    default_channel: :class:`~steam.GroupChannel`
        The group's default channel.
    channels: list[:class:`~steam.GroupChannel`]
        A list of the group's channels.
    """

    __slots__ = (
        "owner",
        "top_members",
        "name",
        "active_member_count",
        "roles",
        "default_role",
        "default_channel",
        "channels",
        "_state",
    )

    def __init__(self, state: ConnectionState, proto: GroupProto):
        super().__init__(proto.chat_group_id, type="Chat")
        self._state = state
        self._from_proto(proto)

    async def __ainit__(self) -> None:
        self.owner = await self._state.client.fetch_user(self.owner)
        self.top_members = await self._state.client.fetch_users(*self.top_members)

    def _from_proto(self, proto: GroupProto) -> None:
        self.owner: User = proto.accountid_owner
        self.name: Optional[str] = proto.chat_group_name or None

        self.active_member_count = proto.active_member_count
        self.top_members: list[User] = proto.top_members
        self.roles: list[Role] = []
        self.default_role: Optional[Role]

        for role in proto.role_actions:
            self.roles.append(Role(self._state, self, role))

        default_role = [r for r in self.roles if r.id == int(proto.default_role_id)]
        self.default_role: Optional[Role] = default_role[0] if default_role else None
        self.channels: list[GroupChannel] = [
            GroupChannel(state=self._state, group=self, channel=channel) for channel in proto.chat_rooms
        ]
        default_channel = [c for c in self.channels if c.id == int(proto.default_chat_id)]
        self.default_channel: Optional[GroupChannel] = default_channel[0] if default_channel else None

    def __repr__(self) -> str:
        attrs = (
            "name",
            "id",
            "owner",
        )
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        return f"<Group {' '.join(resolved)}>"

    def __str__(self) -> str:
        return self.name or ""

    async def leave(self) -> None:
        """|coro|
        Leaves the :class:`Group`.
        """
        await self._state.leave_chat(self.id)

    async def invite(self, user: User) -> None:
        """|coro|
        Invites a :class:`~steam.User` to the :class:`Group`.

        Parameters
        -----------
        user: :class:`~steam.User`
            The user to invite to the group.
        """
        await self._state.invite_user_to_group(user.id64, self.id)



# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2015-2020 Rapptz
Copyright (c) 2020 James

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

EnumMeta from https://github.com/Rapptz/discord.py/blob/master/discord/enums.py
"""

from __future__ import annotations

from types import MappingProxyType
from typing import TYPE_CHECKING, Any, Generator, Mapping, NoReturn, TypeVar, Union

from typing_extensions import Literal

__all__ = (
    "Enum",
    "IntEnum",
    "EResult",
    "EUniverse",
    "EType",
    "ETypeChar",
    "EInstanceFlag",
    "EFriendRelationship",
    "EPersonaState",
    "EPersonaStateFlag",
    "ECommunityVisibilityState",
    "ETradeOfferState",
    "EChatEntryType",
    "EUIMode",
    "EUserBadge",
)

T = TypeVar("T")
E = TypeVar("E", bound="Enum")
IE = TypeVar("IE", bound="IntEnum")


def _is_descriptor(obj: Any) -> bool:
    """Returns True if obj is a descriptor, False otherwise."""
    return hasattr(obj, "__get__") or hasattr(obj, "__set__") or hasattr(obj, "__delete__")


class EnumMeta(type):
    def __new__(mcs, name: str, bases: tuple[type, ...], attrs: dict[str, Any]) -> EnumMeta:
        set_attribute = super().__setattr__
        enum_class = super().__new__(mcs, name, bases, attrs)
        enum_class_new = enum_class.__new__

        value_mapping: dict[Any, Enum] = {}
        member_mapping: dict[str, Enum] = {}

        for key, value in attrs.items():
            if key[0] == "_" or _is_descriptor(value):
                continue

            member = value_mapping.get(value)
            if member is None:
                member = enum_class_new(enum_class, name=key, value=value)
                value_mapping[value] = member

            member_mapping[key] = member
            set_attribute(enum_class, key, member)

        set_attribute(enum_class, "_value_map_", value_mapping)
        set_attribute(enum_class, "_member_map_", member_mapping)
        return enum_class

    def __call__(cls: type[E], value: Any) -> E:
        if value.__class__ is cls:
            return value
        try:
            return cls._value_map_[value]
        except (KeyError, TypeError):
            raise ValueError(f"{value!r} is not a valid {cls.__name__}")

    def __repr__(cls) -> str:
        return f"<enum {cls.__name__!r}>"

    def __iter__(cls: type[E]) -> Generator[E, None, None]:
        yield from cls._member_map_.values()

    def __reversed__(cls: type[E]) -> Generator[E, None, None]:
        yield from reversed(tuple(cls._member_map_.values()))  # can remove tuple cast after 3.7

    def __len__(cls) -> int:
        return len(cls._member_map_)

    def __getitem__(cls: type[E], key: str) -> E:
        return cls._member_map_[key]

    def __setattr__(cls, name: str, value: Any) -> NoReturn:
        raise AttributeError(f"{cls.__name__}: cannot reassign Enum members.")

    def __delattr__(cls, name: str) -> NoReturn:
        raise AttributeError(f"{cls.__name__}: cannot delete Enum members.")

    def __contains__(cls: type[E], member: E) -> bool:
        if not isinstance(member, Enum):
            raise TypeError(
                f"unsupported operand type(s) for 'in': {member.__class__.__qualname__!r} and {cls.__qualname__!r}"
            )
        return isinstance(member, cls) and member.name in cls._member_map_

    @property
    def __members__(cls: type[E]) -> Mapping[str, E]:
        return MappingProxyType(cls._member_map_)


class Enum(metaclass=EnumMeta):
    """A general enumeration, emulates `enum.Enum`."""

    def __new__(cls, *, name: str, value: Any) -> Enum:
        # N.B. this method is not ever called after enum creation as it is shadowed by EnumMeta.__call__ and is just
        # for creating Enum members
        self = super().__new__(cls)
        self.name = name
        self.value = value
        return self

    def __bool__(self) -> Literal[True]:
        return True  # an enum member with a zero value would return False otherwise

    def __str__(self) -> str:
        return f"{self.__class__.__name__}.{self.name}"

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}.{self.name}: {self.value!r}>"

    @classmethod
    def try_value(cls: type[E], value: T) -> Union[E, T]:
        try:
            return cls._value_map_[value]
        except (KeyError, TypeError):
            return value


class IntEnum(Enum, int):
    """An enumeration where all the values are integers, emulates `enum.IntEnum`."""

    def __new__(cls, *, name: str, value: int) -> IntEnum:
        self = int.__new__(cls, value)
        self.name = name
        self.value = value
        return self


if TYPE_CHECKING:
    from enum import Enum as _Enum

    class Enum(_Enum):
        @classmethod
        def try_value(cls: type[E], value: T) -> Union[E, T]:
            ...

    class IntEnum(int, Enum):
        pass

    # pretending these are enum.IntEnum subclasses makes things much nicer for linters as IntEnums have custom behaviour
    # I can't seem to replicate


# fmt: off
class EResult(IntEnum):
    Invalid                         = 0  #: Invalid EResult.
    OK                              = 1  #: Success.
    Fail                            = 2  #: Generic failure.
    NoConnection                    = 3  #: Your Steam client doesn't have a connection to the back-end.
    InvalidPassword                 = 5  #: Password/ticket is invalid.
    LoggedInElsewhere               = 6  #: Same user logged in elsewhere.
    InvalidProtocolVersion          = 7  #: Protocol version is incorrect.
    InvalidParameter                = 8  #: A parameter is incorrect.
    FileNotFound                    = 9  #: File was not found.
    Busy                            = 10  #: Called method busy - action not taken.
    InvalidState                    = 11  #: Called object was in an invalid state.
    InvalidName                     = 12  #: The name was invalid.
    InvalidEmail                    = 13  #: The email was invalid.
    DuplicateName                   = 14  #: The name is not unique.
    AccessDenied                    = 15  #: Access is denied.
    Timeout                         = 16  #: Operation timed out.
    Banned                          = 17  #: VAC2 banned.
    AccountNotFound                 = 18  #: Account not found.
    InvalidSteamID                  = 19  #: The Steam ID was invalid.
    ServiceUnavailable              = 20  #: The requested service is currently unavailable.
    NotLoggedOn                     = 21  #: The user is not logged on.
    Pending                         = 22  #: Request is pending (may be in process, or waiting on third party).
    EncryptionFailure               = 23  #: Encryption or decryption failed.
    InsufficientPrivilege           = 24  #: Insufficient privilege.
    LimitExceeded                   = 25  #: Too much of a good thing.
    Revoked                         = 26  #: Access has been revoked (used for revoked guest passes).
    Expired                         = 27  #: License/Guest pass the user is trying to access is expired.
    AlreadyRedeemed                 = 28  #: Guest pass has already been redeemed by account, cannot be acknowledged again.
    DuplicateRequest                = 29  #: The request is a duplicate, ignored this time.
    AlreadyOwned                    = 30  #: All the games in guest pass redemption request are already owned by the user.
    IPNotFound                      = 31  #: IP address not found.
    PersistFailed                   = 32  #: Failed to write change to the data store.
    LockingFailed                   = 33  #: Failed to acquire access lock for this operation.
    LogonSessionReplaced            = 34  #: The logon session has been replaced.
    ConnectFailed                   = 35  #: Failed to connect.
    HandshakeFailed                 = 36  #: The authentication handshake has failed.
    IOFailure                       = 37  #: Generic IO failure.
    RemoteDisconnect                = 38  #: The remote server has disconnected.
    ShoppingCartNotFound            = 39  #: Failed to find the shopping cart requested.
    Blocked                         = 40  #: A user blocked the action.
    Ignored                         = 41  #: The target is ignoring sender.
    NoMatch                         = 42  #: Nothing matching the request found.
    AccountDisabled                 = 43  #: The account is disabled.
    ServiceReadOnly                 = 44  #: This service is not accepting content changes right now.
    AccountNotFeatured              = 45  #: Account doesn't have value, so this feature isn't available.
    AdministratorOK                 = 46  #: Allowed to take this action, but only because requester is admin.
    ContentVersion                  = 47  #: A Version mismatch in content transmitted within the Steam protocol.
    TryAnotherCM                    = 48  #: The current CM can't service the user making a request, should try another.
    PasswordRequiredToKickSession   = 49  #: You are already logged in elsewhere, this cached credential login has failed.
    AlreadyLoggedInElsewhere        = 50  #: You are already logged in elsewhere, you must wait.
    Suspended                       = 51  #: Long running operation (content download) suspended/paused.
    Cancelled                       = 52  #: Operation canceled (typically by user content download).
    DataCorruption                  = 53  #: Operation canceled because data is malformed or unrecoverable.
    DiskFull                        = 54  #: Operation canceled - not enough disk space.
    RemoteCallFailed                = 55  #: An remote call or IPC call failed.
    ExternalAccountUnlinked         = 57  #: External account is not linked to a Steam account.
    PSNTicketInvalid                = 58  #: PSN ticket was invalid.
    ExternalAccountAlreadyLinked    = 59  #: External account is already linked to some other account.
    RemoteFileConflict              = 60  #: The sync cannot resume due to a conflict between the local and remote files.
    IllegalPassword                 = 61  #: The requested new password is not legal.
    SameAsPreviousValue             = 62  #: New value is the same as the old one (secret question and answer).
    AccountLogonDenied              = 63  #: Account login denied due to 2nd factor authentication failure.
    CannotUseOldPassword            = 64  #: The requested new password is not legal.
    InvalidLoginAuthCode            = 65  #: Account login denied due to auth code invalid.
    AccountLogonDeniedNoMail        = 66  #: Account login denied due to 2nd factor authentication failure.
    HardwareNotCapableOfIPT         = 67  #: The users hardware does not support Intel's identity protection technology.
    IPTInitError                    = 68  #: Intel's Identity Protection Technology has failed to initialize.
    ParentalControlRestricted       = 69  #: Operation failed due to parental control restrictions for current user.
    FacebookQueryError              = 70  #: Facebook query returned an error.
    ExpiredLoginAuthCode            = 71  #: Account login denied due to auth code expired.
    IPLoginRestrictionFailed        = 72  #: The login failed due to an IP restriction.
    AccountLockedDown               = 73  #: The current users account is currently locked for use.
    VerifiedEmailRequired           = 74  #: The logon failed because the accounts email is not verified.
    NoMatchingURL                   = 75  #: There is no url matching the provided values.
    BadResponse                     = 76  #: Parse failure, missing field, etc.
    RequirePasswordReEntry          = 77  #: The user cannot complete the action until they re-enter their password.
    ValueOutOfRange                 = 78  #: The value entered is outside the acceptable range.
    UnexpectedError                 = 79  #: Something happened that we didn't expect to ever happen.
    Disabled                        = 80  #: The requested service has been configured to be unavailable.
    InvalidCEGSubmission            = 81  #: The set of files submitted to the CEG server are not valid.
    RestrictedDevice                = 82  #: The device being used is not allowed to perform this action.
    RegionLocked                    = 83  #: The action could not be complete because it is region restricted.
    RateLimitExceeded               = 84  #: Temporary rate limit exceeded. Different from :attr:`LimitExceeded`.
    LoginDeniedNeedTwoFactor        = 85  #: Need two-factor code to login.
    ItemDeleted                     = 86  #: The thing we're trying to access has been deleted.
    AccountLoginDeniedThrottle      = 87  #: Login attempt failed, try to throttle response to possible attacker.
    TwoFactorCodeMismatch           = 88  #: Two factor code mismatch.
    TwoFactorActivationCodeMismatch = 89  #: Activation code for two-factor didn't match.
    NotModified                     = 91  #: Data not modified.
    TimeNotSynced                   = 93  #: The time presented is out of range or tolerance.
    SMSCodeFailed                   = 94  #: SMS code failure (no match, none pending, etc.).
    AccountActivityLimitExceeded    = 96  #: Too many changes to this account.
    PhoneActivityLimitExceeded      = 97  #: Too many changes to this phone.
    RefundToWallet                  = 98  #: Cannot refund to payment method, must use wallet.
    EmailSendFailure                = 99  #: Cannot send an email.
    NotSettled                      = 100  #: Can't perform operation till payment has settled.
    NeedCaptcha                     = 101  #: Needs to provide a valid captcha.
    GSLTDenied                      = 102  #: A game server login token owned by this token's owner has been banned.
    GSOwnerDenied                   = 103  #: Game server owner is denied for other reason.
    InvalidItemType                 = 104  #: The type of thing we were requested to act on is invalid.
    IPBanned                        = 105  #: The IP address has been banned from taking this action.
    GSLTExpired                     = 106  #: This token has expired from disuse; can be reset for use.
    InsufficientFunds               = 107  #: User doesn't have enough wallet funds to complete the action.
    TooManyPending                  = 108  #: There are too many of this thing pending already.
    NoSiteLicensesFound             = 109  #: No site licenses found.
    WGNetworkSendExceeded           = 110  #: The WG couldn't send a response because we exceeded max network send size.
    AccountNotFriends               = 111  #: Not friends with the relevant account.
    LimitedUserAccount              = 112  #: The account is limited and cannot perform this action.
    CantRemoveItem                  = 113  #: Cannot remove the item.
    AccountHasBeenDeleted           = 114  #: The relevant account has been deleted.
    AccountHasCancelledLicense      = 115  #: The user has a user cancelled license.


class EUniverse(IntEnum):
    Invalid  = 0  #: Invalid.
    Public   = 1  #: The standard public universe.
    Beta     = 2  #: Beta universe used inside Valve.
    Internal = 3  #: Internal universe used inside Valve.
    Dev      = 4  #: Dev universe used inside Valve.
    Max      = 6  #: Total number of universes, used for sanity checks.


class EType(IntEnum):
    Invalid        = 0   #: Used for invalid Steam IDs.
    Individual     = 1   #: Single user account.
    Multiseat      = 2   #: Multiseat (e.g. cybercafe) account.
    GameServer     = 3   #: Game server account.
    AnonGameServer = 4   #: Anonymous game server account.
    Pending        = 5   #: Pending.
    ContentServer  = 6   #: Valve internal content server account.
    Clan           = 7   #: Steam clan.
    Chat           = 8   #: Steam group chat or lobby.
    ConsoleUser    = 9   #: Fake SteamID for local PSN account on PS3 or Live account on 360, etc.
    AnonUser       = 10  #: Anonymous user account. (Used to create an account or reset a password)
    Max            = 11  #: Max of 16 items in this field


class ETypeChar(IntEnum):
    I = EType.Invalid         #: The character used for :class:`~steam.EType.Invalid`.
    U = EType.Individual      #: The character used for :class:`~steam.EType.Individual`.
    M = EType.Multiseat       #: The character used for :class:`~steam.EType.Multiseat`.
    G = EType.GameServer      #: The character used for :class:`~steam.EType.GameServer`.
    A = EType.AnonGameServer  #: The character used for :class:`~steam.EType.AnonGameServer`.
    P = EType.Pending         #: The character used for :class:`~steam.EType.Pending`.
    C = EType.ContentServer   #: The character used for :class:`~steam.EType.ContentServer`.
    g = EType.Clan            #: The character used for :class:`~steam.EType.Clan`.
    T = EType.Chat            #: The character used for :class:`~steam.EType.Chat` (Lobby/group chat).
    L = EType.Chat            #: The character used for :class:`~steam.EType.Chat` (Lobby/group chat).
    c = EType.Clan            #: The character used for :class:`~steam.EType.Clan`.
    a = EType.AnonUser        #: The character used for :class:`~steam.EType.Invalid`.


class EInstanceFlag(IntEnum):
    MMSLobby = 0x20000  #: The Steam ID is for a MMS Lobby.
    Lobby    = 0x40000  #: The Steam ID is for a Lobby.
    Clan     = 0x80000  #: The Steam ID is for a Clan.


class EFriendRelationship(IntEnum):
    NONE             = 0  #: The user has no relationship to you.
    Blocked          = 1  #: The user ignored the invite.
    RequestRecipient = 2  #: The user has requested to be you.
    Friend           = 3  #: The user is friends with you.
    RequestInitiator = 4  #: You have requested to be friends with the user.
    Ignored          = 5  #: You have explicitly blocked this other user from comments/chat/etc.
    IgnoredFriend    = 6  #: The user has ignored the current user.
    Max              = 8  #: The total number of friend relationships used for looping and verification.


class EPersonaState(IntEnum):
    Offline        = 0  #: The user is not currently logged on.
    Online         = 1  #: The user is logged on.
    Busy           = 2  #: The user is on, but busy.
    Away           = 3  #: The user has been marked as AFK for a short period of time.
    Snooze         = 4  #: The user has been marked as AFK for a long period of time.
    LookingToTrade = 5  #: The user is online and wanting to trade.
    LookingToPlay  = 6  #: The user is online and wanting to play.
    Invisible      = 7  #: The user is invisible.
    Max            = 8  #: The total number of states. Only used for looping and validation.


class EPersonaStateFlag(IntEnum):
    NONE                 = 0
    HasRichPresence      = 1
    InJoinableGame       = 2
    Golden               = 4
    RemotePlayTogether   = 8
    ClientTypeWeb        = 256
    ClientTypeMobile     = 512
    ClientTypeTenfoot    = 1024
    ClientTypeVR         = 2048
    LaunchTypeGamepad    = 4096
    LaunchTypeCompatTool = 8192

    @classmethod
    def components(cls, flag: int) -> list[EPersonaStateFlag]:
        """A helper function to breakdown a flag into its component parts.

        Parameters
        ----------
        flag: :class:`int`
            The flag to break down.

        Returns
        -------
        list[:class:`EPersonaStateFlag`]
            The resolved flags.
        """
        flags = [enum for value, enum in cls._value_map_.items() if value & flag]
        value = 0
        for f in flags:
            value |= f
        return flags if value == flag else []


class ECommunityVisibilityState(IntEnum):
    NONE        = 0  #: The user has no community state.
    Private     = 1  #: The user has a private profile.
    FriendsOnly = 2  #: The user has a friends only profile.
    Public      = 3  #: The user has a public profile.


class ETradeOfferState(IntEnum):
    Invalid                   = 1   #: The trade offer's state is invalid.
    Active                    = 2   #: The trade offer is active.
    Accepted                  = 3   #: The trade offer has been accepted.
    Countered                 = 4   #: The trade offer has been countered.
    Expired                   = 5   #: The trade offer has expired.
    Canceled                  = 6   #: The trade offer has been cancelled.
    Declined                  = 7   #: The trade offer has be declined by the partner.
    InvalidItems              = 8   #: The trade offer has invalid items and has been cancelled.
    ConfirmationNeed          = 9   #: The trade offer needs confirmation.
    CanceledBySecondaryFactor = 10  #: The trade offer was cancelled by second factor.
    StateInEscrow             = 11  #: The trade offer is in escrow.


class EChatEntryType(IntEnum):
    Invalid          = 0   #: An Invalid Chat entry.
    Text             = 1   #: A Normal text message from another user.
    Typing           = 2   #: Another user is typing (not used in multi-user chat).
    InviteGame       = 3   #: An Invite from other user into that users current game.
    LeftConversation = 6   #: A user has left the conversation.
    Entered          = 7   #: A user has entered the conversation (used in multi-user chat and group chat).
    WasKicked        = 8   #: A user was kicked.
    WasBanned        = 9   #: A user was banned.
    Disconnected     = 10  #: A user disconnected.
    HistoricalChat   = 11  #: A chat message from user's chat history or offline message.
    LinkBlocked      = 14  #: A link was removed by the chat filter.


class EUIMode(IntEnum):
    Desktop    = 0  #: The UI mode for the desktop client.
    BigPicture = 1  #: The UI mode for big picture mode.
    Mobile     = 2  #: The UI mode for mobile.
    Web        = 3  #: The UI mode for the web client.


class EUserBadge(IntEnum):
    Invalid                           = 0  #: Invalid Badge.
    YearsOfService                    = 1  #: The years of service badge.
    Community                         = 2  #: The pillar of the community badge.
    Portal2PotatoARG                  = 3  #: The portal to potato badge.
    TreasureHunt                      = 4  #: The treasure hunter badge.
    SummerSale2011                    = 5  #: The Summer sale badge for 2011.
    WinterSale2011                    = 6  #: The Winter sale badge for 2011.
    SummerSale2012                    = 7  #: The Summer sale badge for 2012.
    WinterSale2012                    = 8  #: The Winter sale badge for 2012.
    CommunityTranslator               = 9  #: The community translator badge.
    CommunityModerator                = 10  #: The community moderator badge.
    ValveEmployee                     = 11  #: The Valve employee badge.
    GameDeveloper                     = 12  #: The game developer badge.
    GameCollector                     = 13  #: The game collector badge.
    TradingCardBetaParticipant        = 14  #: The trading card beta participant badge.
    SteamBoxBeta                      = 15  #: The Steam box beta badge.
    Summer2014RedTeam                 = 16  #: The Summer sale badge for 2014 for the red team.
    Summer2014BlueTeam                = 17  #: The Summer sale badge for 2014 for the blue team.
    Summer2014PinkTeam                = 18  #: The Summer sale badge for 2014 for the pink team.
    Summer2014GreenTeam               = 19  #: The Summer sale badge for 2014 for the green team.
    Summer2014PurpleTeam              = 20  #: The Summer sale badge for 2014 for the purple team.
    Auction2014                       = 21  #: The auction badge for 2014.
    GoldenProfile2014                 = 22  #: The golden profile for 2014.
    TowerAttackMiniGame               = 23  #: The tower attack mini-game badge.
    Winter2015ARGRedHerring           = 24  #: The Winter ARG red herring badge for 2015.
    SteamAwards2016Nominations        = 25  #: The Steam Awards Nominations badge for 2016.
    StickerCompletionist2017          = 26  #: The sticker completionist badge for 2017.
    SteamAwards2017Nominations        = 27  #: The Steam Awards Nominations badge for 2017.
    SpringCleaning2018                = 28  #: The Spring cleaning badge for 2018.
    Salien                            = 29  #: The Salien badge.
    RetiredModerator                  = 30  #: The retired moderator badge.
    SteamAwards2018Nominations        = 31  #: The Steam Awards Nominations badge for 2018.
    ValveModerator                    = 32  #: The Valve moderator badge.
    WinterSale2018                    = 33  #: The Winter sale badge for 2018.
    LunarNewYearSale2019              = 34  #: The lunar new years sale badge for 2019.
    LunarNewYearSale2019GoldenProfile = 35  #: The lunar new year golden profile sale badge for 2019.
    SpringCleaning2019                = 36  #: The Spring cleaning badge for 2019.
    Summer2019                        = 37  #: The Summer sale badge for 2019.
    Summer2019TeamHare                = 38  #: The Summer sale badge for 2014 for team hare.
    Summer2019TeamTortoise            = 39  #: The Summer sale badge for 2014 for team tortoise.
    Summer2019TeamCorgi               = 40  #: The Summer sale badge for 2014 for team corgi.
    Summer2019TeamCockatiel           = 41  #: The Summer sale badge for 2014 for team cockatiel.
    Summer2019TeamPig                 = 42  #: The Summer sale badge for 2014 for team pig.
    SteamAwards2019Nominations        = 43  #: The Steam Awards Nominations badge for 2019.
    WinterSaleEvent2019               = 44  #: The Winter sale badge for 2019.



# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2020 James

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

from datetime import datetime

import attr

from .enums import EUserBadge
from .game import Game

__all__ = (
    "Badge",
    "FavouriteBadge",
    "UserBadges",
)


@attr.dataclass(slots=True)
class FavouriteBadge:
    name: str
    xp: int = attr.ib(converter=int)  # it's passed as a str
    description: str
    icon_url: str
    level: int = attr.ib(converter=int)  # it's passed as a str


class Badge:
    """Represents a Steam badge.

    Attributes
    ----------
    id: Union[:class:`.EUserBadge`, :class:`int`]
        The badge's ID.
    level: :class:`int`
        The badge's level.
    xp: :class:`int`
        The badge's XP.
    completion_time: :class:`datetime.datetime`
        The time the badge was completed at.
    scarcity: :class:`int`
        The scarcity of the badge.
    game: Optional[:class:`~steam.Game`]
        The game associated with the badge.
    """

    __slots__ = ("id", "xp", "game", "level", "scarcity", "completion_time")

    def __init__(self, data: dict):
        self.id = EUserBadge.try_value(data["badgeid"])
        self.level: int = data["level"]
        self.xp: int = data["xp"]
        self.completion_time = datetime.utcfromtimestamp(data["completion_time"])
        self.scarcity: int = data["scarcity"]
        self.game = Game(id=data["appid"]) if "appid" in data else None

    def __repr__(self) -> str:
        attrs = ("id", "level", "xp", "game")
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        return f'<Badge {" ".join(resolved)}>'


class UserBadges:
    """Represents a Steam :class:`~steam.User`'s badges/level.

    Attributes
    ----------
    level: :class:`int`
        The badge's level.
    xp: :class:`int`
        The badge's XP.
    xp_needed_to_level_up: :class:`int`
        The amount of XP the user needs to level up.
    xp_needed_for_current_level: :class:`int`
        The amount of XP the user's current level requires to achieve.
    badges: list[:class:`Badge`]
        A list of the user's badges.
    """

    __slots__ = (
        "xp",
        "level",
        "badges",
        "xp_needed_to_level_up",
        "xp_needed_for_current_level",
    )

    def __init__(self, data: dict):
        self.level: int = data["player_level"]
        self.xp: int = data["player_xp"]
        self.xp_needed_to_level_up: int = data["player_xp_needed_to_level_up"]
        self.xp_needed_for_current_level: int = data["player_xp_needed_current_level"]
        self.badges: list[Badge] = [Badge(data) for data in data["badges"]]

    def __repr__(self) -> str:
        attrs = ("level", "xp")
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        return f'<UserBadges {" ".join(resolved)}>'

    def __len__(self) -> int:
        return len(self.badges)



# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2020 James

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

import asyncio
import inspect
import re
from datetime import datetime
from typing import TYPE_CHECKING, Optional, Union

from bs4 import BeautifulSoup

from . import utils
from .abc import Commentable, SteamID
from .channel import ClanChannel
from .errors import HTTPException
from .game import Game
from .models import community_route
from .protobufs.steammessages_chat import (
    CChatRoomSummaryPair as ReceivedResponse,
    CClanChatRoomsGetClanChatRoomInfoResponse as FetchedResponse,
)
from .role import Role

if TYPE_CHECKING:
    from .state import ConnectionState
    from .user import User

__all__ = ("Clan",)


class Clan(Commentable, comment_path="Clan"):
    """Represents a Steam clan.

    .. container:: operations

        .. describe:: x == y

            Checks if two clans are equal.

        .. describe:: str(x)

            Returns the clan's name.


    Attributes
    ------------
    name: :class:`str`
        The name of the clan.
    chat_id: :class:`int`
        The clan's chat_id.
    url: :class:`str`
        The url of the clan.
    icon_url: :class:`str`
        The icon url of the clan. Uses the large (184x184 px) image url.
    description: :class:`str`
        The description of the clan.
    tagline: Optional[:class:`str`]
        The clan's tagline.
    member_count: :class:`int`
        The amount of users in the clan.
    online_count: :class:`int`
        The amount of users currently online.
    active_member_count: :class:`int`
        The amount of currently users in the clan's chat room.
    in_game_count: :class:`int`
        The amount of user's currently in game.
    created_at: :class:`datetime.datetime`
        The time the clan was created_at.
    language: :class:`str`
        The language set for the clan.
    location: :class:`str`
        The location set for the clan.
    game: :class:`.Game`
        The clan's associated game.
    owner: :class:`~steam.User`
        The clan's owner.
    admins: list[:class:`~steam.User`]
        A list of the clan's administrators.
    mods: list[:class:`~steam.User`]
        A list of the clan's moderators.
    top_members: list[:class:`~steam.User`]
        A list of the clan's top_members.
    roles: list[:class:`.Role`]
        A list of the clan's roles.
    default_role: :class:`.Role`
        The clan's default_role.
    channels: list[:class:`.ClanChannel`]
        A list of the clan's channels.
    default_channel: :class:`.ClanChannel`
        The clan's default_channel.
    """

    __slots__ = (
        "url",
        "name",
        "description",
        "icon_url",
        "created_at",
        "language",
        "location",
        "member_count",
        "in_game_count",
        "online_count",
        "admins",
        "mods",
        "chat_id",
        "active_member_count",
        "owner",
        "default_role",
        "tagline",
        "top_members",
        "roles",
        "channels",
        "default_channel",
        "game",
    )

    # TODO more to implement https://github.com/DoctorMcKay/node-steamcommunity/blob/master/components/clans.js

    def __init__(self, state: ConnectionState, id: int):
        super().__init__(id, type="Clan")
        self.url = community_route(f"gid/{id}")
        self._state = state
        self.name: Optional[str] = None
        self.chat_id: Optional[int] = None
        self.tagline: Optional[str] = None
        self.game: Optional[Game] = None
        self.owner: Optional[User] = None
        self.active_member_count: Optional[int] = None
        self.top_members: list[User] = []
        self.roles: list[Role] = []
        self.default_role: Optional[Role] = None
        self.channels: list[ClanChannel] = []
        self.default_channel: Optional[ClanChannel] = None

    async def __ainit__(self) -> None:
        resp = await self._state.request("GET", self.url)
        if not self.id:
            search = re.search(r"OpenGroupChat\(\s*'(\d+)'\s*\)", resp)
            if search is None:
                return
            super().__init__(search.group(1), type="Clan")

        soup = BeautifulSoup(resp, "html.parser")
        self.name = soup.find("title").text[28:]
        self.description = soup.find("meta", attrs={"property": "og:description"})["content"]
        self.icon_url = soup.find("link", attrs={"rel": "image_src"})["href"]
        stats = soup.find("div", attrs={"class": "grouppage_resp_stats"})
        for stat in stats.find_all("div", attrs={"class": "groupstat"}):
            if "Founded" in stat.text:
                text = stat.text.split("Founded")[1].strip()
                if ", " not in stat.text:
                    text = f"{text}, {datetime.utcnow().year}"
                try:
                    self.created_at = datetime.strptime(text, "%d %B, %Y" if text.split()[0].isdigit() else "%B %d, %Y")
                except ValueError:  # why do other countries have to exist
                    self.created_at = None
            if "Language" in stat.text:
                self.language = stat.text.split("Language")[1].strip()
            if "Location" in stat.text:
                self.location = stat.text.split("Location")[1].strip()

        for count in stats.find_all("div", attrs={"class": "membercount"}):
            if "MEMBERS" in count.text:
                self.member_count = int(count.text.split("MEMBERS")[0].strip().replace(",", ""))
            if "IN-GAME" in count.text:
                self.in_game_count = int(count.text.split("IN-GAME")[0].strip().replace(",", ""))
            if "ONLINE" in count.text:
                self.online_count = int(count.text.split("ONLINE")[0].strip().replace(",", ""))

        admins = []
        mods = []
        is_admins = True
        for fields in soup.find_all("div", attrs={"class": "membergrid"}):
            for idx, field in enumerate(fields.find_all("div")):
                if "Members" in field.text:
                    if mods:
                        mods.pop()
                    break
                if "Moderators" in field.text:
                    officer = admins.pop()
                    mods.append(officer)
                    is_admins = False
                try:
                    account_id = fields.find_all("div", attrs={"class": "playerAvatar"})[idx]["data-miniprofile"]
                except IndexError:
                    break
                else:
                    if is_admins:
                        admins.append(account_id)
                    else:
                        mods.append(account_id)

        self.admins = await self._state.client.fetch_users(*admins)
        self.mods = await self._state.client.fetch_users(*mods)

    @classmethod
    async def _from_proto(cls, state: ConnectionState, clan_proto: Union[ReceivedResponse, FetchedResponse]) -> Clan:
        if isinstance(clan_proto, ReceivedResponse):
            id = clan_proto.group_summary.clanid
        else:
            id = clan_proto.chat_group_summary.clanid
        self = cls(state, id)
        await self.__ainit__()
        if isinstance(clan_proto, ReceivedResponse):
            proto = clan_proto.group_summary
        else:
            proto = clan_proto.chat_group_summary

        self.chat_id = proto.chat_group_id
        self.tagline = proto.chat_group_tagline or None
        self.active_member_count = proto.active_member_count
        self.game = Game(id=proto.appid)

        self.owner = await self._state.fetch_user(utils.make_id64(proto.accountid_owner))
        self.top_members = await self._state.fetch_users([utils.make_id64(u) for u in proto.top_members])

        self.roles = []
        for role in proto.role_actions:
            self.roles.append(Role(self._state, self, role))
        try:
            self.default_role = [r for r in self.roles if r.id == int(proto.default_role_id)][0]
        except IndexError:
            pass

        self.channels = []
        self.default_channel = None
        if isinstance(clan_proto, ReceivedResponse):
            channels = clan_proto.user_chat_group_state.user_chat_room_state
        else:
            return self
        for channel in channels:
            channel = ClanChannel(state=self._state, clan=self, channel=channel)
            if channel not in self.channels:
                self.channels.append(channel)
            else:  # update the old instance
                idx = self.channels.index(channel)
                old_channel = self.channels[idx]
                for name, attr in inspect.getmembers(channel):
                    if attr:
                        setattr(old_channel, name, attr)

        default_channel = [c for c in self.channels if c.id == int(proto.default_chat_id)]
        self.default_channel = default_channel[0] if default_channel else None
        return self

    def __repr__(self) -> str:
        attrs = ("name", "id", "chat_id", "type", "universe", "instance")
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        return f"<Clan {' '.join(resolved)}>"

    def __str__(self) -> str:
        return self.name

    async def fetch_members(self) -> list[SteamID]:
        """|coro|
        Fetches a clan's member list.

        Returns
        --------
        list[:class:`~steam.SteamID`]
            A basic list of the clan's members.
            This can be a very slow operation due to
            the rate limits on this endpoint.
        """
        ret = []
        resp = await self._state.request("GET", f"{self.url}/members?p=1&content_only=true")
        soup = BeautifulSoup(resp, "html.parser")
        pages = int(soup.find_all("a", attrs={"class": "pagelink"}).pop().text)

        async def getter(i: int) -> None:
            try:
                resp = await self._state.request("GET", f"{self.url}/members?p={i + 1}")
            except HTTPException:
                await asyncio.sleep(20)
                await getter(i)
            else:
                soup = BeautifulSoup(resp, "html.parser")
                for s in soup.find_all("div", attrs={"id": "memberList"}):
                    for user in s.find_all("div", attrs={"class": "member_block"}):
                        ret.append(SteamID(user["data-miniprofile"]))

        for i in range(pages):
            await getter(i)

        return ret

    async def join(self) -> None:
        """|coro|
        Joins the :class:`Clan`. This will also join the clan's chat.
        """
        await self._state.http.join_clan(self.id64)
        await self._state.join_chat(self.chat_id)

    async def leave(self) -> None:
        """|coro|
        Leaves the :class:`Clan`.
        """
        await self._state.http.leave_clan(self.id64)

    async def invite(self, user: User) -> None:
        """|coro|
        Invites a :class:`~steam.User` to the :class:`Clan`.

        Parameters
        -----------
        user: :class:`~steam.User`
            The user to invite to the clan.
        """
        await self._state.http.invite_user_to_clan(user_id64=user.id64, clan_id=self.id64)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: steammessages_useraccount.proto
# plugin: python-betterproto

from dataclasses import dataclass
from typing import List

import betterproto


class EInternalAccountType(betterproto.Enum):
    SteamAccountType = 1
    ClanType = 2
    AppType = 3
    BroadcastChannelType = 4


class EExternalAccountType(betterproto.Enum):
    NONE = 0
    SteamAccount = 1
    GoogleAccount = 2
    FacebookAccount = 3
    TwitterAccount = 4
    TwitchAccount = 5
    YouTubeChannelAccount = 6
    FacebookPage = 7


@dataclass(eq=False, repr=False)
class CUserAccountGetAvailableValveDiscountPromotionsRequest(betterproto.Message):
    country_code: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class CUserAccountGetAvailableValveDiscountPromotionsResponse(betterproto.Message):
    promotions: List[
        "CUserAccountGetAvailableValveDiscountPromotionsResponseValveDiscountPromotionDetails"
    ] = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class CUserAccountGetAvailableValveDiscountPromotionsResponseValveDiscountPromotionDetails(betterproto.Message):
    promotionid: int = betterproto.uint32_field(1)
    promotion_description: str = betterproto.string_field(2)
    minimum_cart_amount: int = betterproto.int64_field(3)
    minimum_cart_amount_for_display: int = betterproto.int64_field(4)
    discount_amount: int = betterproto.int64_field(5)
    currency_code: int = betterproto.int32_field(6)
    available_use_count: int = betterproto.int32_field(7)
    promotional_discount_type: int = betterproto.int32_field(8)
    loyalty_reward_id: int = betterproto.int32_field(9)
    localized_name_token: str = betterproto.string_field(10)
    max_use_count: int = betterproto.int32_field(11)


@dataclass(eq=False, repr=False)
class CUserAccountGetAccountLinkStatusRequest(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class CUserAccountGetAccountLinkStatusResponse(betterproto.Message):
    pwid: int = betterproto.uint32_field(1)
    identity_verification: int = betterproto.uint32_field(2)
    performed_age_verification: bool = betterproto.bool_field(3)


@dataclass(eq=False, repr=False)
class CUserAccountCancelLicenseForAppRequest(betterproto.Message):
    appid: int = betterproto.uint32_field(1)


@dataclass(eq=False, repr=False)
class CUserAccountCancelLicenseForAppResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class CUserAccountCreateFriendInviteTokenRequest(betterproto.Message):
    invite_limit: int = betterproto.uint32_field(1)
    invite_duration: int = betterproto.uint32_field(2)
    invite_note: str = betterproto.string_field(3)


@dataclass(eq=False, repr=False)
class CUserAccountCreateFriendInviteTokenResponse(betterproto.Message):
    invite_token: str = betterproto.string_field(1)
    invite_limit: int = betterproto.uint64_field(2)
    invite_duration: int = betterproto.uint64_field(3)
    time_created: int = betterproto.fixed32_field(4)
    valid: bool = betterproto.bool_field(5)


@dataclass(eq=False, repr=False)
class CUserAccountGetFriendInviteTokensRequest(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class CUserAccountGetFriendInviteTokensResponse(betterproto.Message):
    tokens: List["CUserAccountCreateFriendInviteTokenResponse"] = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class CUserAccountViewFriendInviteTokenRequest(betterproto.Message):
    steamid: int = betterproto.fixed64_field(1)
    invite_token: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class CUserAccountViewFriendInviteTokenResponse(betterproto.Message):
    valid: bool = betterproto.bool_field(1)
    steamid: int = betterproto.uint64_field(2)
    invite_duration: int = betterproto.uint64_field(3)


@dataclass(eq=False, repr=False)
class CUserAccountRedeemFriendInviteTokenRequest(betterproto.Message):
    steamid: int = betterproto.fixed64_field(1)
    invite_token: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class CUserAccountRedeemFriendInviteTokenResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class CUserAccountRevokeFriendInviteTokenRequest(betterproto.Message):
    invite_token: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class CUserAccountRevokeFriendInviteTokenResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class CUserAccountRegisterCompatToolRequest(betterproto.Message):
    compat_tool: int = betterproto.uint32_field(1)


@dataclass(eq=False, repr=False)
class CUserAccountRegisterCompatToolResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class CAccountLinkingGetLinkedAccountInfoRequest(betterproto.Message):
    account_type: "EInternalAccountType" = betterproto.enum_field(1)
    account_id: int = betterproto.uint64_field(2)
    filter: "EExternalAccountType" = betterproto.enum_field(3)
    return_access_token: bool = betterproto.bool_field(4)


@dataclass(eq=False, repr=False)
class CAccountLinkingGetLinkedAccountInfoResponse(betterproto.Message):
    external_accounts: List[
        "CAccountLinkingGetLinkedAccountInfoResponseCExternalAccountTupleResponse"
    ] = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class CAccountLinkingGetLinkedAccountInfoResponseCExternalAccountTupleResponse(betterproto.Message):
    external_type: "EExternalAccountType" = betterproto.enum_field(1)
    external_id: str = betterproto.string_field(2)
    external_user_name: str = betterproto.string_field(3)
    external_url: str = betterproto.string_field(4)
    access_token: str = betterproto.string_field(5)
    access_token_secret: str = betterproto.string_field(6)
    is_valid: bool = betterproto.bool_field(7)


@dataclass(eq=False, repr=False)
class CEmbeddedClientAuthorizeCurrentDeviceRequest(betterproto.Message):
    steamid: int = betterproto.fixed64_field(1)
    appid: int = betterproto.uint32_field(2)
    device_info: str = betterproto.string_field(3)
    deviceid: int = betterproto.uint32_field(4)


@dataclass(eq=False, repr=False)
class CEmbeddedClientToken(betterproto.Message):
    steamid: int = betterproto.fixed64_field(1)
    client_token: bytes = betterproto.bytes_field(2)
    expiry: int = betterproto.uint32_field(3)
    deviceid: int = betterproto.uint32_field(4)


@dataclass(eq=False, repr=False)
class CEmbeddedClientAuthorizeDeviceResponse(betterproto.Message):
    result: int = betterproto.uint32_field(1)
    token: "CEmbeddedClientToken" = betterproto.message_field(2)
